# IMPORT LIBRARIES
from .N_GUI import GUI
from .N_Label import Label
from .N_Custom import Event_Bind

class Button(Label):

    def __init__(self, Main, *Args, **Kwargs):
        self._GUI = GUI._Instance
        if self._GUI is not None:
            self._Type = "Button"
            try:
                super().__init__(Main, *Args, **Kwargs)
                self._Config = ['Name', 'Auto_Dark', 'Background', 'Light_Background', 'Dark_Background', 'Foreground', 'Light_Foreground', 'Dark_Foreground', 'Border_Color', 'Light_Border_Color', 'Dark_Border_Color', 'Border_Size', 'Resize', 'Popup', 'Display', 'Left', 'Top', 'Width', 'Height', 'Animate_Left', 'Animate_Top', 'Animate_Width', 'Animate_Height', 'Animate_Time', 'Font_Size', 'Font_Weight', 'Font_Family', 'Align', 'Value', 'Hover_Background', 'Light_Hover_Background', 'Dark_Hover_Background', 'Hover_Foreground', 'Light_Hover_Foreground', 'Dark_Hover_Foreground', 'Hover_Border_Color', 'Light_Hover_Border_Color', 'Dark_Hover_Border_Color', 'Disable']
                self._Type = "Button"
                self._Disable = False
                Event_Bind(self._Widget, On_Click=lambda E: self.On_Click(E), On_Release=lambda E: self.On_Release(E))
            except Exception as E:
                self._GUI.Error(f"{self._Type} -> Init -> {E}")
        else:
            print("Error: Gluonix -> GUI Instance Has Not Been Created")

    def __str__(self):
        return "Nucleon_Glunoix_Button[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Button[]"
            
    def Bind(self, **Input):
        try:
            if 'On_Show' in Input:
                self._On_Show = Input['On_Show']
            if 'On_Hide' in Input:
                self._On_Hide = Input['On_Hide']
            if 'On_Animate' in Input:
                self._On_Animate = Input['On_Animate']
            if 'On_Change' in Input:
                self._On_Change = Input['On_Change']
            Event_Bind(self._Widget, **Input)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind -> {E}")
            
    def On_Click(self, E):
        try:
            if self._Disable or not self._Widget:
                return
            self._Widget.config(background=self._GUI.Fade(self._Background, 70), foreground=self._Foreground)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Press -> {E}")
            
    def On_Release(self, E):
        try:
            if self._Disable or not self._Widget:
                return
            self._Widget.config(background=self._Background, foreground=self._Foreground)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Release -> {E}")


Event_Map = {
    'On_Click': '<Button-1>',
    'On_Release': '<ButtonRelease-1>',
    'On_Double_Click': '<Double-Button-1>',
    'On_Triple_Click': '<Triple-Button-1>',
    'On_Right_Click': '<Button-3>',
    'On_Right_Release': '<ButtonRelease-3>',
    'On_Right_Double_Click': '<Double-Button-3>',
    'On_Right_Triple_Click': '<Triple-Button-3>',
    'On_Middle_Click': '<Button-2>',
    'On_Middle_Release': '<ButtonRelease-2>',
    'On_Middle_Double_Click': '<Double-Button-2>',
    'On_Middle_Triple_Click': '<Triple-Button-2>',
    'On_Drag': '<B1-Motion>',
    'On_Right_Drag': '<B3-Motion>',
    'On_Middle_Drag': '<B2-Motion>',
    'On_Hover_In': '<Enter>',
    'On_Hover_Out': '<Leave>',
    'On_Motion': '<Motion>',
    'On_Focus_In': '<FocusIn>',
    'On_Focus_Out': '<FocusOut>',
    'On_Key': '<KeyPress>',
    'On_Key_Release': '<KeyRelease>',
    'On_Configure': '<Configure>',
    'On_Map': '<Map>',
    'On_Unmap': '<Unmap>',
    'On_Destroy': '<Destroy>',
    'On_Expose': '<Expose>',
    'On_Visibility': '<Visibility>',
    'On_Mouse_Wheel': ['<MouseWheel>', '<Button-4>', '<Button-5>'],
    'On_Copy': '<<Copy>>',
    'On_Paste': '<<Paste>>',
    'On_Cut': '<<Cut>>',
    'On_Undo': '<<Undo>>',
    'On_Redo': '<<Redo>>',
    'On_Control_Click': '<Control-Button-1>',
    'On_Control_Release': '<Control-ButtonRelease-1>',
    'On_Control_Double_Click': '<Control-Double-Button-1>',
    'On_Control_Triple_Click': '<Control-Triple-Button-1>',
    'On_Control_Right_Click': '<Control-Button-3>',
    'On_Control_Right_Release': '<Control-ButtonRelease-3>',
    'On_Control_Right_Double_Click': '<Control-Double-Button-3>',
    'On_Control_Right_Triple_Click': '<Control-Triple-Button-3>',
    'On_Control_Middle_Click': '<Control-Button-2>',
    'On_Control_Middle_Release': '<Control-ButtonRelease-2>',
    'On_Control_Middle_Double_Click': '<Control-Double-Button-2>',
    'On_Control_Middle_Triple_Click': '<Control-Triple-Button-2>',
    'On_Control_Drag': '<Control-B1-Motion>',
    'On_Control_Right_Drag': '<Control-B3-Motion>',
    'On_Control_Middle_Drag': '<Control-B2-Motion>',
    'On_Control_Hover_In': '<Control-Enter>',
    'On_Control_Hover_Out': '<Control-Leave>',
    'On_Control_Mouse_Wheel': ['<Control-MouseWheel>', '<Control-Button-4>', '<Control-Button-5>'],
    'On_Shift_Click': '<Shift-Button-1>',
    'On_Shift_Release': '<Shift-ButtonRelease-1>',
    'On_Shift_Double_Click': '<Shift-Double-Button-1>',
    'On_Shift_Triple_Click': '<Shift-Triple-Button-1>',
    'On_Shift_Right_Click': '<Shift-Button-3>',
    'On_Shift_Right_Release': '<Shift-ButtonRelease-3>',
    'On_Shift_Right_Double_Click': '<Shift-Double-Button-3>',
    'On_Shift_Right_Triple_Click': '<Shift-Triple-Button-3>',
    'On_Shift_Middle_Click': '<Shift-Button-2>',
    'On_Shift_Middle_Release': '<Shift-ButtonRelease-2>',
    'On_Shift_Middle_Double_Click': '<Shift-Double-Button-2>',
    'On_Shift_Middle_Triple_Click': '<Shift-Triple-Button-2>',
    'On_Shift_Drag': '<Shift-B1-Motion>',
    'On_Shift_Right_Drag': '<Shift-B3-Motion>',
    'On_Shift_Middle_Drag': '<Shift-B2-Motion>',
    'On_Shift_Hover_In': '<Shift-Enter>',
    'On_Shift_Hover_Out': '<Shift-Leave>',
    'On_Shift_Mouse_Wheel': ['<Shift-MouseWheel>', '<Shift-Button-4>', '<Shift-Button-5>'],
    'On_Alt_Click': '<Alt-Button-1>',
    'On_Alt_Release': '<Alt-ButtonRelease-1>',
    'On_Alt_Double_Click': '<Alt-Double-Button-1>',
    'On_Alt_Triple_Click': '<Alt-Triple-Button-1>',
    'On_Alt_Right_Click': '<Alt-Button-3>',
    'On_Alt_Right_Release': '<Alt-ButtonRelease-3>',
    'On_Alt_Right_Double_Click': '<Alt-Double-Button-3>',
    'On_Alt_Right_Triple_Click': '<Alt-Triple-Button-3>',
    'On_Alt_Middle_Click': '<Alt-Button-2>',
    'On_Alt_Middle_Release': '<Alt-ButtonRelease-2>',
    'On_Alt_Middle_Double_Click': '<Alt-Double-Button-2>',
    'On_Alt_Middle_Triple_Click': '<Alt-Triple-Button-2>',
    'On_Alt_Drag': '<Alt-B1-Motion>',
    'On_Alt_Right_Drag': '<Alt-B3-Motion>',
    'On_Alt_Middle_Drag': '<Alt-B2-Motion>',
    'On_Alt_Hover_In': '<Alt-Enter>',
    'On_Alt_Hover_Out': '<Alt-Leave>',
    'On_Alt_Mouse_Wheel': ['<Alt-MouseWheel>', '<Alt-Button-4>', '<Alt-Button-5>'],
}

def Event_Bind(Widget, **Input):
    Bind_Add = "+"
    if 'Add' in Input and Input['Add'] is False:
        Bind_Add = ""
    if 'Cursor' in Input:
        Widget.config(cursor=Input['Cursor'])
    elif 'Cursor_Hand' in Input and Input['Cursor_Hand']:
        Widget.config(cursor="hand2")
    elif 'Cursor_Loading' in Input and Input['Cursor_Loading']:
        Widget.config(cursor="watch")
    elif 'Cursor_Resize_Vertical' in Input and Input['Cursor_Resize_Vertical']:
        Widget.config(cursor='sb_v_double_arrow')
    elif 'Cursor_Resize_Horizontal' in Input and Input['Cursor_Resize_Horizontal']:
        Widget.config(cursor='sb_h_double_arrow')
    elif 'Cursor_Arrow' in Input and Input['Cursor_Arrow']:
        Widget.config(cursor='arrow')
    for Key, Value in Input.items():
        if Key in Event_Map:
            Sequences = Event_Map[Key]
            if isinstance(Sequences, list):
                for Seq in Sequences:
                    Widget.bind(Seq, Value, add=Bind_Add)
            else:
                Widget.bind(Sequences, Value, add=Bind_Add)

def Event_Bind_Canvas(Widget, Item, **Input):
    Bind_Add = "+"
    if 'Add' in Input and Input['Add'] is False:
        Bind_Add = ""
    for Key, Value in Input.items():
        if Key in Event_Map:
            Sequences = Event_Map[Key]
            if isinstance(Sequences, list):
                for Seq in Sequences:
                    Widget.tag_bind(Item, Seq, Value, add=Bind_Add)
            else:
                Widget.tag_bind(Item, Sequences, Value, add=Bind_Add)


# IMPORT MAIN LIBRARIES
import os
import sys
import _thread
import time
import datetime
import PIL
import io
import base64
import sqlite3
import urllib.request
import tkinter as TK
from tkinter import font
from tkinter import filedialog
from tkinter import scrolledtext
from tkinter import ttk

# IMPORT CUSTOM GUI LIBRARIES
from .N_GUI import GUI
from .N_Popup import Popup
from .N_Frame import Frame
from .N_Group import Group
from .N_Canvas import Canvas
from .N_Canvas_Line import Canvas_Line
from .N_Canvas_Polyline import Canvas_Polyline
from .N_Canvas_Pie import Canvas_Pie
from .N_Canvas_Arc import Canvas_Arc
from .N_Canvas_Circle import Canvas_Circle
from .N_Canvas_Oval import Canvas_Oval
from .N_Canvas_Rectangle import Canvas_Rectangle
from .N_Canvas_RectangleR import Canvas_RectangleR
from .N_Canvas_Polygon import Canvas_Polygon
from .N_Canvas_Image import Canvas_Image
from .N_Canvas_Dynamic_Image import Canvas_Dynamic_Image
from .N_Canvas_Text import Canvas_Text
from .N_Canvas_Text_Old import Canvas_Text_Old
from .N_Scroll import Scroll
from .N_Separator import Separator
from .N_Label import Label
from .N_Roubel import Roubel
from .N_Button import Button
from .N_Entry import Entry
from .N_Line import Line
from .N_Text import Text
from .N_Tree import Tree
from .N_Bar import Bar
from .N_Spinner import Spinner
from .N_Scale import Scale
from .N_Select import Select
from .N_List import List
from .N_Video import Video
from .N_Player import Player
from .N_Editor import Editor
from .N_Image import Image
from .N_Dynamic_Image import Dynamic_Image
from .N_Compound import Compound
from .N_Check import Check
from .N_Switch import Switch
from .N_Radio import Radio
from .N_Radio import Variable


# IMPORT CUSTOM OTHER LIBRARIES
from .N_SQL import SQL


# IMPORT LIBRARIES
import os
from io import BytesIO
import urllib
from PIL import Image as PIL_Image, ImageTk as PIL_ImageTk, ImageMath as PIL_ImageMath
import tkinter as TK
import threading, math, time
from .N_GUI import GUI
from .N_Custom import Event_Bind

PIL_Image.MAX_IMAGE_PIXELS = None

class Image:
    
    def __init__(self, Main, *args, **kwargs):
        self._GUI = GUI._Instance
        if self._GUI is not None:
            self._Type = "Image"
            try:
                self._Config = ['Name', 'Auto_Dark', 'Background', 'Light_Background', 'Dark_Background', 'Use_Foreground', 'Foreground', 'Light_Foreground', 'Dark_Foreground', 'Border_Color', 'Light_Border_Color', 'Dark_Border_Color', 'Border_Size', 'Resize', 'Popup', 'Display', 'Left', 'Top', 'Width', 'Height', 'Animate_Left', 'Animate_Top', 'Animate_Width', 'Animate_Height', 'Animate_Time', 'Path', 'Path_Initial', 'Rotate', 'Transparent', 'Aspect_Ratio', 'Convert_Type', 'Tolerance', 'Hover_Background', 'Light_Hover_Background', 'Dark_Hover_Background', 'Hover_Foreground', 'Light_Hover_Foreground', 'Dark_Hover_Foreground', 'Hover_Border_Color', 'Light_Hover_Border_Color', 'Dark_Hover_Border_Color']
                self._Initialized = False
                self._Name = False
                self._Last_Name = False
                self._Resize = True
                self._Popup = False
                self._Display = True
                self._Size_Update = False
                self._Main = Main
                self._Widget = TK.Label(self._Main._Frame)
                self._Background = self._Main._Background
                self._Background_Main = True
                self._Border_Color = '#000000'
                self._Border_Size = 0
                self._Use_Foreground = False
                self._Foreground = False
                self._Hover_Background = False
                self._Hover_Foreground = False
                self._Hover_Border_Color = False
                self._Last_Background = False
                self._Last_Foreground = False
                self._Last_Border_Color = False
                self._Animating = False
                self._Anim_Stop = None
                self._Anim_Thread = None
                self._Animate_Ease = lambda t: (1 - (1 - t)**3)
                self._Animate_Speed = None
                self._Animate_Left = 0
                self._Animate_Top = 0
                self._Animate_Width = 0
                self._Animate_Height = 0
                self._Animate_Time = 1.0
                self._Tolerance = 10
                self._Image = False
                self._Image_Width = 0
                self._Image_Height = 0
                self._Path = False
                self._Path_Memory = False
                self._Path_Initial = False
                self._Rotate = 0
                self._Angle = 0
                self._Transparent = True
                self._Aspect_Ratio = True
                self._Convert_Type = 'RGBA'
                self._Auto_Dark = True
                self._On_Show = False
                self._On_Hide = False
                self._On_Animate = False
                self._Is_Gif = False
                self._Gif_Frames = []
                self._Gif_Durations = []
                self._Gif_Loop = 1
                self._Gif_Index = 0
                self._Gif_Stop = threading.Event()
                self._Gif_Thread = None
                self._Gif_Running = False
                self._TK_Image = None
                self._Configure_After_Id = None
                self._Configure_Width = 0
                self._Configure_Height = 0
                self._Last_Width = 0
                self._Last_Height = 0
            except Exception as E:
                self._GUI.Error(f"{self._Type} -> Init -> {E}")
        else:
            print("Error: Gluonix -> GUI Instance Has Not Been Created")
    
    def __str__(self):
        return "Nucleon_Glunoix_Image[]"
    
    def __repr__(self):
        return "Nucleon_Glunoix_Image[]"
    
    def Copy(self, Name=False, Main=False):
        try:
            if not Main:
                Main = self._Main
            Instance = type(self)(Main)
            for Key in self._Config:
                if hasattr(self, "_"+Key):
                    setattr(Instance, "_"+Key, getattr(self, "_"+Key))
            setattr(Instance, "_Name", Name)
            Instance.Create()
            return Instance
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Copy -> {E}")
    
    def Delete(self):
        try:
            self.Animate_Cancel()
            self.Stop()
            self._Main._Widget.remove(self)
            self._Widget.destroy()
            if self:
                del self
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete -> {E}")
    
    def Hide(self):
        try:
            self._Display = False
            self.Animate_Cancel()
            self.Stop()
            self._Widget.place_forget()
            if self._On_Hide:
                self._On_Hide()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Hide -> {E}")
    
    def _Build_Place_Args(self, Left, Top, Width, Height):
        try:
            Place_Args = {}
            Parent_Width = float(getattr(self._Main, "_Width", 0) or 0)
            Parent_Height = float(getattr(self._Main, "_Height", 0) or 0)
            if self._Resize and self._Main._Type!='Scroll' and Parent_Width > 0 and Parent_Height > 0:
                Place_Args["relx"] = float(Left) / Parent_Width
                Place_Args["rely"] = float(Top) / Parent_Height
                if Width>0:
                    Place_Args["relwidth"] = float(Width) / Parent_Width
                if Height>0:
                    Place_Args["relheight"] = float(Height) / Parent_Height
            else:
                Place_Args["x"] = int(round(Left))
                Place_Args["y"] = int(round(Top))
                if Width>0:
                    Place_Args["width"] = int(round(Width))
                if Height>0:
                    Place_Args["height"] = int(round(Height))
            return Place_Args
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Build_Place_Args -> {E}")
            return {}

    def _Place_Geometry(self, Left, Top, Width, Height):
        try:
            Place_Args = self._Build_Place_Args(Left, Top, Width, Height)
            self._Widget.place(**Place_Args)
            self._Widget.lift()
            if self._Image and len(Place_Args)>2:
                self.Load()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Place_Geometry -> {E}")
    
    def Show(self):
        try:
            if self._Animating:
                return
            self._Display = True
            self._Place_Geometry(self._Left, self._Top, self._Width, self._Height)
            if self._On_Show:
                self._On_Show()
            self.Run()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Show -> {E}")
    
    def Grab(self, Path=False):
        try:
            return self._GUI.Grab_Widget(Path=Path, Widget=self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Grab -> {E}")
    
    def Animate(self, Hide=False, Thread=True):
        try:
            self.Animate_Cancel()
            Final_Left = float(getattr(self, "_Left", 0))
            Final_Top = float(getattr(self, "_Top", 0))
            Final_Width = float(getattr(self, "_Width", 0))
            Final_Height = float(getattr(self, "_Height", 0))
            Start_Left = float(self._Animate_Left)
            Start_Top = float(self._Animate_Top)
            Animate_Width = float(getattr(self, "_Animate_Width", 0) or 0)
            Animate_Height = float(getattr(self, "_Animate_Height", 0) or 0)
            Size_Anim = not (Animate_Width == 0 and Animate_Height == 0)
            Start_Width = Animate_Width if Size_Anim else Final_Width
            Start_Height = Animate_Height if Size_Anim else Final_Height
            Same_Pos = int(round(Start_Left)) == int(round(Final_Left)) and int(round(Start_Top)) == int(round(Final_Top))
            Same_Size = int(round(Start_Width)) == int(round(Final_Width)) and int(round(Start_Height)) == int(round(Final_Height))
            if Same_Pos and (not Size_Anim or Same_Size):
                def Snap_Same():
                    if not self._Widget.winfo_exists():
                        return
                    self._Place_Geometry(Final_Left, Final_Top, Final_Width, Final_Height)
                    self._Animating = False
                    if Hide:
                        self.Hide()
                    if self._On_Animate:
                        self._On_Animate()
                self._GUI._Frame.after(0, Snap_Same)
                return
            def Show_Start():
                if not self._Widget.winfo_exists():
                    return
                self._Place_Geometry(Start_Left, Start_Top, Start_Width, Start_Height)
            self._GUI._Frame.after(0, Show_Start)
            Dx = Final_Left - Start_Left
            Dy = Final_Top - Start_Top
            Dw = Final_Width - Start_Width if Size_Anim else 0.0
            Dh = Final_Height - Start_Height if Size_Anim else 0.0
            Dist = math.hypot(math.hypot(Dx, Dy), math.hypot(Dw, Dh))
            if Dist == 0.0:
                def Snap_Final():
                    if not self._Widget.winfo_exists():
                        return
                    self._Place_Geometry(Final_Left, Final_Top, Final_Width, Final_Height)
                    self._Animating = False
                    if Hide:
                        self.Hide()
                    if self._On_Animate:
                        self._On_Animate()
                self._GUI._Frame.after(0, Snap_Final)
                return
            if self._Animate_Speed and self._Animate_Speed > 0:
                Duration = max(0.001, Dist / float(self._Animate_Speed))
            else:
                Duration = max(0.001, float(self._Animate_Time))
            Ease = self._Animate_Ease or (lambda T: T)
            Target_FPS = 90.0
            Frame_Interval = 1.0 / Target_FPS
            self._Animating = True
            Stop = threading.Event()
            self._Anim_Stop = Stop
            def Worker():
                T0 = time.perf_counter()
                Next_Tick = T0
                Last = None
                while not Stop.is_set():
                    Now = time.perf_counter()
                    T = (Now - T0) / Duration
                    if T >= 1.0:
                        def Snap_End():
                            if Stop.is_set() or not self._Widget.winfo_exists():
                                return
                            self._Place_Geometry(Final_Left, Final_Top, Final_Width, Final_Height)
                            self._Animating = False
                            if Hide:
                                self.Hide()
                            if self._On_Animate:
                                self._On_Animate()
                        self._GUI._Frame.after(0, Snap_End)
                        return
                    K = Ease(max(0.0, min(1.0, T)))
                    X = Start_Left + Dx * K
                    Y = Start_Top + Dy * K
                    W = Start_Width + Dw * K
                    H = Start_Height + Dh * K
                    Cur = (int(round(X)), int(round(Y)), int(round(W)), int(round(H)))
                    if Cur != Last:
                        Last = Cur
                        def Post(C=Cur):
                            if Stop.is_set() or not self._Widget.winfo_exists():
                                return
                            if self._Animating:
                                self._Place_Geometry(C[0], C[1], C[2], C[3])
                        self._GUI._Frame.after(0, Post)
                    Next_Tick += Frame_Interval
                    Sleep_For = Next_Tick - time.perf_counter()
                    if Sleep_For < -2 * Frame_Interval:
                        Next_Tick = time.perf_counter()
                        Sleep_For = Frame_Interval
                    if Sleep_For > 0:
                        if Stop.wait(Sleep_For):
                            return
            self.Show()
            if Thread:
                T = threading.Thread(target=Worker, daemon=True)
                self._Anim_Thread = T
                T.start()
            else:
                Worker()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate -> {E}")
            self.Animate_Cancel()
    
    def Animate_Cancel(self):
        try:
            self._Animating = False
            if self._Anim_Stop is not None:
                self._Anim_Stop.set()
            self._Anim_Stop = None
            self._Anim_Thread = None
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate_Cancel -> {E}")

    def Clear(self):
        try:
            if self._Image:
                try:
                    self._Image.close()
                except:
                    pass
            self._Image = False
            self._Image_Width = 0
            self._Image_Height = 0
            self._Widget.configure(image=None)
            self._TK_Image = None
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Clear -> {E}")

    def Set(self, Path):
        try:
            self.Stop()
            self._Path = Path
            self._Path_Memory = self._Path
            if isinstance(Path, str):
                Parsed = urllib.parse.urlparse(Path)
                if Parsed.scheme in ("http", "https") or self.Is_Large_File(Path):
                    self.Async_Open(Path)
                    return
            self.Open()
            self.Load()
            if self._Display and not self._Animating:
                self.Show()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Set -> {E}")

    def Initial(self):
        try:
            if self._Path_Initial:
                if hasattr(self, 'Stop'):
                    self.Stop()
                self.Set(self._Path_Initial)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Initial -> {E}")
    
    def Reset(self):
        try:
            self._Angle = 0
            self.Open()
            self.Load()
            if self._Display and not self._Animating:
                self.Show()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Reset -> {E}")
            
    def Refresh(self):
        try:
            self.Set(Path=self._Path)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Refresh -> {E}")
    
    def Widget(self):
        try:
            return self._Widget
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Widget -> {E}")
    
    def Bind(self, **Input):
        try:
            if 'On_Show' in Input:
                self._On_Show = Input['On_Show']
            if 'On_Hide' in Input:
                self._On_Hide = Input['On_Hide']
            if 'On_Animate' in Input:
                self._On_Animate = Input['On_Animate']
            Event_Bind(self._Widget, **Input)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind -> {E}")
    
    def On_Hover_In(self, E):
        try:
            Config = {}
            if self._Hover_Background:
                self._Last_Background = self._Background
                Config['Background'] = self._Hover_Background
            if self._Hover_Foreground:
                self._Last_Foreground = self._Foreground
                Config['Foreground'] = self._Hover_Foreground
            if self._Hover_Border_Color:
                self._Last_Border_Color = self._Border_Color
                Config['Border_Color'] = self._Hover_Border_Color
            if len(Config)>0:
                self.Config(**Config)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Hover_In -> {E}")
    
    def On_Hover_Out(self, E):
        try:
            Config = {}
            if self._Hover_Background and self._Last_Background:
                Config['Background'] = self._Last_Background if self._Background==self._Hover_Background else self._Background
            if self._Hover_Foreground and self._Last_Foreground:
                Config['Foreground'] = self._Last_Foreground if self._Foreground==self._Hover_Foreground else self._Foreground
            if self._Hover_Border_Color and self._Last_Border_Color:
                Config['Border_Color'] = self._Last_Border_Color if self._Border_Color==self._Hover_Border_Color else self._Border_Color
            if len(Config)>0:
                self.Config(**Config)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Hover_Out -> {E}")
    
    def Config_Get(self, *Input):
        try:
            Return = {}
            for Each in self._Config:
                if Each in Input:
                    Return[Each] = getattr(self, "_"+Each)
            return Return
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config_Get -> {E}")
    
    def Config(self, **Input):
        try:
            Run = False
            for Each in self._Config:
                if Each in Input:
                    Value = Input[Each]
                    setattr(self, "_"+Each, Value)
                    Run = True
            if self._Initialized and Run:
                self.Create()
            if "Background" in Input:
                self._Background_Main = not bool(Input["Background"])
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config -> {E}")
    
    def Move(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left += Left
            if Top is not None:
                self._Top += Top
            if (Left is not None or Top is not None) and self._Initialized and self._Display and not self._Animating:
                self.Position(Left=self._Left, Top=self._Top)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Move -> {E}")
    
    def Center(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left-self._Width/2
            if Top is not None:
                self._Top = Top-self._Height/2
            if (Left is not None or Top is not None) and self._Initialized and self._Display and not self._Animating:
                self.Position(Left=self._Left, Top=self._Top)
            return [self._Left+self._Width/2, self._Top+self._Height/2]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Center -> {E}")
    
    def Position(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left
            if Top is not None:
                self._Top = Top
            if (Left is not None or Top is not None) and self._Initialized and self._Display and not self._Animating:
                self._Place_Geometry(self._Left, self._Top, 0, 0)
            return [int(self._Left), int(self._Top)]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Position -> {E}")
    
    def Size(self, Width=False, Height=False):
        try:
            if Width:
                self._Width = Width
            if Height:
                self._Height = Height
            if (Width or Height) and self._Initialized and self._Display and not self._Animating:
                if self._Display and not self._Animating:
                    self._Place_Geometry(self._Left, self._Top, self._Width, self._Height)
            return [int(self._Width), int(self._Height)]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Size -> {E}")
    
    def Enlarge(self, Value=None):
        try:
            if Value is not None:
                self._Left -= Value
                self._Top -= Value
                self._Width += Value
                self._Height += Value
                if self._Initialized and self._Display and not self._Animating:
                    self._Place_Geometry(self._Left, self._Top, self._Width, self._Height)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Enlarge -> {E}")
    
    def Shrink(self, Value=None):
        try:
            if Value is not None:
                self._Left += Value
                self._Top += Value
                self._Width -= Value
                self._Height -= Value
                if self._Initialized and self._Display and not self._Animating:
                    self._Place_Geometry(self._Left, self._Top, self._Width, self._Height)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Shrink -> {E}")
            
    def Box(self):
        try:
            return [self._Widget.winfo_x(), self._Widget.winfo_y(), self._Widget.winfo_width(), self._Widget.winfo_height()]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Box -> {E}")
        
    def Ratio(self):
        try:
            Width_Ratio = self._Widget.winfo_width()/self._Width
            Height_Ratio = self._Widget.winfo_height()/self._Height
            return [Width_Ratio, Height_Ratio]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Ratio -> {E}")
    
    def Locate(self, Width, Height, Left, Top):
        try:
            Width = self._Width*(Width/100)
            Height = self._Height*(Height/100)
            Left = self._Width*(Left/100)
            Top = self._Height*(Top/100)
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate -> {E}")
    
    def Locate_Reverse(self, Width, Height, Left, Top):
        try:
            Width = round((Width/self._Width)*100, 3)
            Height = round((Height/self._Height)*100, 3)
            Left = round((Left/self._Width)*100, 3)
            Top = round((Top/self._Height)*100, 3)
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate_Reverse -> {E}")
    
    def Create(self):
        try:
            if not self._Background:
                self._Background = self._Main._Background
                if not hasattr(self, "_Light_Background"):
                    setattr(self, "_Light_Background", self._Background)
                if not hasattr(self, "_Dark_Background"):
                    setattr(self, "_Dark_Background", self._GUI.Invert(self._Background))
            if self._Auto_Dark and not self._GUI._Dark_Mode:
                self.Update_Color()
            if not self._Initialized:
                self.Update_Color()
                if not self._Display:
                    self.Hide()
                self._Main._Widget.append(self)
                Event_Bind(self._Widget, On_Hover_In=lambda E: self.On_Hover_In(E), On_Hover_Out=lambda E: self.On_Hover_Out(E))
                self._Widget.bind("<Configure>", self.On_Configure, add="+")
                self._Initialized = True
            self._Widget.config(background=self._Background)
            self._Widget['highlightbackground']=self._Border_Color
            self._Widget['highlightcolor']=self._Border_Color
            self._Widget['highlightthickness']=self._Border_Size
            if isinstance(self._Path, str) and isinstance(self._Path_Memory, str):
                if self._Path != self._Path_Memory:
                    self._Path_Memory = self._Path
                    self.Open()
            elif isinstance(self._Path, list) and isinstance(self._Path_Memory, list):
                if not all(a == b for a, b in zip(self._Path, self._Path_Memory)):
                    self._Path_Memory = self._Path
                    self.Open()
            elif type(self._Path) != type(self._Path_Memory):
                self._Path_Memory = self._Path
                self.Open()
            if self._Display:
                self.Show()
            else:
                self.Hide()
            if self._Name!=self._Last_Name:
                if self._Last_Name:
                    if self._Last_Name in self._Main.__dict__:
                        del self._Main.__dict__[self._Last_Name]
                if self._Name:
                    self._Main.__dict__[self._Name] = self
                self._Last_Name = self._Name
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Create -> {E}")
    
    def Update_Color(self):
        try:
            self._GUI.Initiate_Colors(self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Update_Color -> {E}")
    
    def RGB(self, HEX):
        try:
            if HEX.startswith('#'):
                HEX = HEX[1:]
            R = int(HEX[0:2], 16)
            G = int(HEX[2:4], 16)
            B = int(HEX[4:6], 16)
            return (R, G, B)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> RGB -> {E}")

    def Is_Large_File(self, Path):
        try:
            if isinstance(Path, str) and os.path.exists(Path):
                Threshold_Bytes = 8*1024*1024
                return os.path.getsize(Path) > Threshold_Bytes
            return False
        except Exception:
            return False

    def Async_Open(self, Path):
        try:
            Path_Obj = Path
            def Worker():
                try:
                    Data = None
                    if isinstance(Path_Obj, str):
                        Parsed = urllib.parse.urlparse(Path_Obj)
                        if Parsed.scheme in ("http", "https"):
                            with urllib.request.urlopen(Path_Obj) as Response:
                                Data = Response.read()
                        elif os.path.exists(Path_Obj):
                            with open(Path_Obj, "rb") as File_Handle:
                                Data = File_Handle.read()
                                if not self._Path_Initial:
                                    self._Path_Initial = Path_Obj
                        else:
                            Data = None
                    if Data is None:
                        def Post_Clear():
                            self.Clear()
                        self._GUI._Frame.after(0, Post_Clear)
                        return
                    with PIL_Image.open(BytesIO(Data)) as Pil_Obj:
                        Is_Gif = False
                        Gif_Frames = []
                        Gif_Durations = []
                        Gif_Loop = 1
                        if getattr(Pil_Obj, "is_animated", False):
                            Is_Gif = True
                            try:
                                Gif_Loop = int(getattr(Pil_Obj, "info", {}).get("loop", 1))
                            except:
                                Gif_Loop = 1
                            Frame_Count = getattr(Pil_Obj, "n_frames", 0) or 0
                            Frame_Index = 0
                            while True:
                                try:
                                    Pil_Obj.seek(Frame_Index)
                                    Frame_Image = Pil_Obj.convert("RGBA").copy()
                                    Duration = int(getattr(Pil_Obj, "info", {}).get("duration", 100))
                                    Gif_Frames.append(Frame_Image)
                                    Gif_Durations.append(max(1, Duration))
                                    Frame_Index += 1
                                    if Frame_Count and Frame_Index >= Frame_Count:
                                        break
                                except EOFError:
                                    break
                            if Gif_Frames:
                                Image_Obj = Gif_Frames[0]
                            else:
                                Is_Gif = False
                                Image_Obj = Pil_Obj.copy()
                        else:
                            Image_Obj = Pil_Obj.copy()
                        Image_Width, Image_Height = Image_Obj.size
                    def Post():
                        try:
                            if Path_Obj != self._Path:
                                return
                            self._Is_Gif = Is_Gif
                            self._Gif_Frames = Gif_Frames
                            self._Gif_Durations = Gif_Durations
                            self._Gif_Loop = Gif_Loop
                            self._Gif_Index = 0
                            self._Image = Image_Obj
                            self._Image_Width = Image_Width
                            self._Image_Height = Image_Height
                            self.Load()
                            if self._Display and not self._Animating:
                                self.Show()
                        except Exception as E2:
                            self._GUI.Error(f"{self._Type} -> Async_Open -> {E2}")
                    self._GUI._Frame.after(0, Post)
                except Exception as E3:
                    def Post_Error():
                        self.Clear()
                        self._GUI.Error(f"{self._Type} -> Async_Open -> {E3}")
                    self._GUI._Frame.after(0, Post_Error)
            Thread_Obj = threading.Thread(target=Worker, daemon=True)
            Thread_Obj.start()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Async_Open -> {E}")

    def Open(self):
        try:
            self.Stop()
            self._Is_Gif = False
            self._Gif_Frames = []
            self._Gif_Durations = []
            self._Gif_Loop = 1
            self._Gif_Index = 0
            if self._Image:
                try:
                    self._Image.close()
                except:
                    pass
            self._Image = False
            Path_Obj = self._Path
            if hasattr(Path_Obj, "__class__") and getattr(type(Path_Obj), "__module__", "").startswith("numpy"):
                self._Image = PIL_Image.fromarray(Path_Obj)
                self._Image_Width, self._Image_Height = self._Image.size
                return
            if isinstance(Path_Obj, PIL_Image.Image):
                Pil_Obj = Path_Obj
                if getattr(Pil_Obj, "is_animated", False):
                    self._Is_Gif = True
                    try:
                        self._Gif_Loop = int(getattr(Pil_Obj, "info", {}).get("loop", 1))
                    except:
                        self._Gif_Loop = 1
                    Frame_Count = getattr(Pil_Obj, "n_frames", 0) or 0
                    Frame_Index = 0
                    while True:
                        try:
                            Pil_Obj.seek(Frame_Index)
                            Frame_Image = Pil_Obj.convert("RGBA").copy()
                            Duration = int(getattr(Pil_Obj, "info", {}).get("duration", 100))
                            self._Gif_Frames.append(Frame_Image)
                            self._Gif_Durations.append(max(1, Duration))
                            Frame_Index += 1
                            if Frame_Count and Frame_Index >= Frame_Count:
                                break
                        except EOFError:
                            break
                    if self._Gif_Frames:
                        self._Image = self._Gif_Frames[0]
                        self._Image_Width, self._Image_Height = self._Image.size
                    else:
                        self._Is_Gif = False
                        self._Image = Pil_Obj.copy()
                        self._Image_Width, self._Image_Height = self._Image.size
                else:
                    self._Image = Pil_Obj.copy()
                    self._Image_Width, self._Image_Height = self._Image.size
                return
            Data = None
            if isinstance(Path_Obj, str):
                Parsed = urllib.parse.urlparse(Path_Obj)
                if Parsed.scheme in ("http", "https"):
                    with urllib.request.urlopen(Path_Obj) as Response:
                        Data = Response.read()
                elif os.path.exists(Path_Obj):
                    with open(Path_Obj, "rb") as File_Handle:
                        Data = File_Handle.read()
                    if not self._Path_Initial:
                        self._Path_Initial = Path_Obj
                else:
                    self.Clear()
                    return
            else:
                self.Clear()
                return
            with PIL_Image.open(BytesIO(Data)) as Pil_Obj:
                if getattr(Pil_Obj, "is_animated", False):
                    self._Is_Gif = True
                    try:
                        self._Gif_Loop = int(Pil_Obj.info.get("loop", 1))
                    except:
                        self._Gif_Loop = 1
                    Frame_Count = getattr(Pil_Obj, "n_frames", 0) or 0
                    Frame_Index = 0
                    while True:
                        try:
                            Pil_Obj.seek(Frame_Index)
                            Frame_Image = Pil_Obj.convert("RGBA").copy()
                            Duration = int(Pil_Obj.info.get("duration", 100))
                            self._Gif_Frames.append(Frame_Image)
                            self._Gif_Durations.append(max(1, Duration))
                            Frame_Index += 1
                            if Frame_Count and Frame_Index >= Frame_Count:
                                break
                        except EOFError:
                            break
                    if self._Gif_Frames:
                        self._Image = self._Gif_Frames[0]
                        self._Image_Width, self._Image_Height = self._Image.size
                    else:
                        self._Is_Gif = False
                        self._Image = Pil_Obj.copy()
                        self._Image_Width, self._Image_Height = self._Image.size
                else:
                    self._Image = Pil_Obj.copy()
                    self._Image_Width, self._Image_Height = self._Image.size
        except Exception as E:
            self.Clear()
            self._GUI.Error(f"{self._Type} -> Open -> {E}")

    def Convert(self, Frame_Width, Frame_Height):
        try:
            "Image -> Convert -> height and width must be > 0"
            if not self._Image or Frame_Width<=0 or Frame_Height<=0:
                return False
            Temp_Image = self._Image.rotate(self._Rotate+self._Angle, PIL_Image.NEAREST, expand=0)
            Image_Ratio = self._Image_Width / self._Image_Height
            Frame_Ratio = Frame_Width / Frame_Height
            if Image_Ratio>=Frame_Ratio:
                Width = Frame_Width
                Width_Ratio = Width / self._Image_Width
                Height = self._Image_Height * Width_Ratio
                Top = (Frame_Height - Height) / 2
                Left = 0
            if Image_Ratio<Frame_Ratio:
                Height = Frame_Height
                Height_Ratio = Height / self._Image_Height
                Width = self._Image_Width * Height_Ratio
                Top = 0
                Left = (Frame_Width - Width) / 2
            if self._Transparent and self._Convert_Type=='RGBA' and self._Foreground and self._Use_Foreground:
                Temp_Image = Temp_Image.convert(self._Convert_Type)
                Temp_Color = self.RGB(self._Foreground)
                Pixel_Data = Temp_Image.load()
                Temp_Width, Temp_Height = Temp_Image.size
                for Y in range(Temp_Height):
                    for X in range(Temp_Width):
                        R, G, B, A = Pixel_Data[X, Y]
                        if R == 0 and G == 0 and B == 0:
                            Pixel_Data[X, Y] = (*Temp_Color, A)
            if self._Aspect_Ratio:
                if int(Width)<=0 or int(Height)<=0:
                    return False
                Temp_Image = Temp_Image.resize((int(Width), int(Height)), PIL_Image.NEAREST)
            else:
                Temp_Image = Temp_Image.resize((int(Frame_Width), int(Frame_Height)), PIL_Image.NEAREST)
            Temp_Image_TK = PIL_ImageTk.PhotoImage(Temp_Image)
            return {"Image": Temp_Image_TK, "Top": Top, "Left": Left}
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Convert -> {E}")

    def Load(self):
        try:
            Left, Top, Width, Height = self.Box()
            if Height>0 and Width>0 and self._Image:
                Image = self.Convert(Width-self._Border_Size*2, Height-self._Border_Size*2)
                if Image:
                    self._Widget.configure(image=Image['Image'])
                    self._TK_Image = Image['Image']
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Load -> {E}")
    
    def Rotate(self, Value=0):
        try:
            self._Angle+=Value
            self.Load()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Rotate -> {E}")
    
    def Run(self):
        try:
            if not self._Is_Gif or len(self._Gif_Frames)==0:
                return
            if self._Gif_Running:
                return
            self._Gif_Stop.clear()
            self._Gif_Running = True
            def Worker():
                Loops = 0
                while self._Gif_Running and not self._Gif_Stop.is_set():
                    if not self._Widget.winfo_exists():
                        break
                    Frame = self._Gif_Frames[self._Gif_Index]
                    Dur = self._Gif_Durations[self._Gif_Index]
                    def Post(F=Frame):
                        if not self._Widget.winfo_exists():
                            return
                        self._Image = F
                        self._Image_Width, self._Image_Height = F.size
                        self.Load()
                    self._GUI._Frame.after(0, Post)
                    if self._Gif_Stop.wait(Dur/1000.0):
                        break
                    self._Gif_Index = (self._Gif_Index+1) % len(self._Gif_Frames)
                    if self._Gif_Index==0:
                        Loops += 1
                        if self._Gif_Loop!=0 and Loops>=self._Gif_Loop:
                            break
                self._Gif_Running = False
            self._Gif_Thread = threading.Thread(target=Worker, daemon=True)
            self._Gif_Thread.start()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Run -> {E}")
    
    def Stop(self):
        try:
            if self._Gif_Thread and self._Gif_Thread.is_alive():
                self._Gif_Running = False
                self._Gif_Stop.set()
                self._Gif_Thread.join(timeout=0.2)
            self._Gif_Stop.clear()
            self._Gif_Thread = None
            self._Gif_Index = 0
            if self._Is_Gif and len(self._Gif_Frames)>0:
                self._Image = self._Gif_Frames[0]
                self._Image_Width, self._Image_Height = self._Image.size
                self.Load()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Stop -> {E}")
    
    def On_Configure(self, Event):
        try:
            Width = int(Event.width)
            Height = int(Event.height)
            if self._Last_Width == Width and self._Last_Height == Height:
                return
            self._Last_Width = Width
            self._Last_Height = Height
            self._Configure_Width = Width
            self._Configure_Height = Height
            if self._Configure_After_Id is not None:
                self._Widget.after_cancel(self._Configure_After_Id)
            self._Configure_After_Id = self._Widget.after(50, self.On_Configure_Debounced)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Configure -> {E}")
                
    def On_Configure_Debounced(self):
        try:
            self._Configure_After_Id = None
            Width = self._Configure_Width
            Height = self._Configure_Height
            if Height>0 and Width>0:
                T = threading.Thread(target=self.On_Configure_Thread, args=(Width, Height), daemon=True)
                T.start()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Configure_Debounced -> {E}")
                
    def On_Configure_Thread(self, Width, Height):
        try:
            if Height>0 and Width>0:
                Image = self.Convert(Width-(self._Border_Size*2), Height-(self._Border_Size*2))
                if Image:
                    self._Widget.configure(image=Image['Image'])
                    self._TK_Image = Image['Image']
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Configure_Thread -> {E}")



# IMPORT LIBRARIES
from PIL import Image as PIL_Image, ImageTk as PIL_ImageTk
from .N_Image import Image
from .N_Custom import Event_Bind

PIL_Image.MAX_IMAGE_PIXELS = None
            
class Dynamic_Image(Image):
    
    def __init__(self, Main, *args, **kwargs):
        super().__init__(Main, *args, **kwargs)
        self._Type = "Dynamic_Image"
        self._Zoom_Scale = 1.0
        self._Zoom_Center = None
        self._Drag_Last_X = 0
        self._Drag_Last_Y = 0
        Event_Bind(self._Widget, On_Click=lambda E: self.Drag_Start(E), On_Drag=lambda E: self.Drag(E), On_Mouse_Wheel=lambda E: self.Zoom(E), On_Right_Click=lambda E: self.Reset(E))

    def __str__(self):
        return "Nucleon_Glunoix_Dynamic_Image[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Dynamic_Image[]"
    
    def Clear(self):
        try:
            super().Clear()
            self._Zoom_Scale = 1.0
            self._Zoom_Center = None
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Clear -> {E}")

    def Initial(self):
        try:
            if self._Path_Initial:
                if hasattr(self, 'Stop'):
                    self.Stop()
                self.Set(self._Path_Initial)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Initial -> {E}")
    
    def Refresh(self):
        try:
            self.Open()
            self.Load()
            if self._Display and not self._Animating:
                self.Show()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Refresh -> {E}")
    
    def Open(self):
        try:
            super().Open()
            if self._Image:
                self._Image_Width, self._Image_Height = self._Image.size
                if self._Zoom_Center is None:
                    self._Zoom_Center = (self._Image_Width // 2, self._Image_Height // 2)
                CX, CY = self._Zoom_Center
                CX = max(0, min(CX, self._Image_Width))
                CY = max(0, min(CY, self._Image_Height))
                self._Zoom_Center = (CX, CY)
                if not hasattr(self, "_Zoom_Scale"):
                    self._Zoom_Scale = 1.0
            else:
                self.Clear()
        except Exception as E:
            self.Clear()
            self._GUI.Error(f"{self._Type} -> Open -> {E}")
    
    def Convert(self, Frame_Width, Frame_Height):
        try:
            if not self._Image or Frame_Width<=0 or Frame_Height<=0:
                return False
            if self._Zoom_Center is None:
                self._Zoom_Center = (self._Image_Width // 2, self._Image_Height // 2)
            Zoom_W = self._Image_Width / self._Zoom_Scale
            Zoom_H = self._Image_Height / self._Zoom_Scale
            CX, CY = self._Zoom_Center
            CX = max(Zoom_W / 2, min(self._Image_Width - Zoom_W / 2, CX))
            CY = max(Zoom_H / 2, min(self._Image_Height - Zoom_H / 2, CY))
            self._Zoom_Center = (CX, CY)
            Left = CX - Zoom_W / 2
            Top = CY - Zoom_H / 2
            Right = Left + Zoom_W
            Bottom = Top + Zoom_H
            Crop = self._Image.crop((int(Left), int(Top), int(Right), int(Bottom)))
            Angle = getattr(self, "_Rotate", 0) + getattr(self, "_Angle", 0)
            if Angle:
                Crop = Crop.rotate(Angle, PIL_Image.NEAREST, expand=0)
            if self._Transparent:
                Crop = Crop.convert("RGBA")
            Image_Width, Image_Height = Crop.size
            if Image_Width<=0 or Image_Height<=0:
                return False
            Inner_W = max(1, int(Frame_Width))
            Inner_H = max(1, int(Frame_Height))
            Aspect_Ratio = getattr(self, "_Aspect_Ratio", True)
            Image_Aspect = Image_Width / Image_Height
            Frame_Aspect = Inner_W / Inner_H
            if not Aspect_Ratio:
                New_Width = Inner_W
                New_Height = Inner_H
                Crop = Crop.resize((New_Width, New_Height), PIL_Image.NEAREST)
                Left_Off = 0
                Top_Off = 0
            else:
                Use_Cover = self._Zoom_Scale>1.0001
                if Use_Cover:
                    if Image_Aspect>Frame_Aspect:
                        New_Height = Inner_H
                        New_Width = int(max(1, round(Inner_H*Image_Aspect)))
                    else:
                        New_Width = Inner_W
                        New_Height = int(max(1, round(Inner_W/Image_Aspect)))
                    if New_Width<=0 or New_Height<=0:
                        return False
                    Crop = Crop.resize((New_Width, New_Height), PIL_Image.NEAREST)
                    Left_Off = max(0, (New_Width - Inner_W)//2)
                    Top_Off = max(0, (New_Height - Inner_H)//2)
                    Crop = Crop.crop((Left_Off, Top_Off, Left_Off+Inner_W, Top_Off+Inner_H))
                    Left_Off = 0
                    Top_Off = 0
                else:
                    if Image_Aspect>Frame_Aspect:
                        New_Width = Inner_W
                        New_Height = int(max(1, round(Inner_W/Image_Aspect)))
                    else:
                        New_Height = Inner_H
                        New_Width = int(max(1, round(Inner_H*Image_Aspect)))
                    if New_Width<=0 or New_Height<=0:
                        return False
                    Crop = Crop.resize((New_Width, New_Height), PIL_Image.NEAREST)
                    Left_Off = (Inner_W - New_Width) // 2
                    Top_Off = (Inner_H - New_Height) // 2
            Crop_TK = PIL_ImageTk.PhotoImage(Crop)
            return {"Image": Crop_TK, "Top": Top_Off, "Left": Left_Off}
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Convert -> {E}")
            return False

    def Drag_Start(self, Event):
        try:
            self._Drag_Last_X = Event.x
            self._Drag_Last_Y = Event.y
        except Exception:
            self._Drag_Last_X, self._Drag_Last_Y = 0, 0
    
    def Drag(self, Event):
        try:
            if not self._Image:
                return
            Curr_X = Event.x
            Curr_Y = Event.y
            Delta_X = Curr_X - self._Drag_Last_X
            Delta_Y = Curr_Y - self._Drag_Last_Y
            self._Drag_Last_X = Curr_X
            self._Drag_Last_Y = Curr_Y
            View_W = self._Image_Width / self._Zoom_Scale
            View_H = self._Image_Height / self._Zoom_Scale
            Inner_W = max(1, self._Width - 2 * self._Border_Size)
            Inner_H = max(1, self._Height - 2 * self._Border_Size)
            Move_X_Img = -Delta_X * (View_W / Inner_W)
            Move_Y_Img = -Delta_Y * (View_H / Inner_H)
            New_CX = self._Zoom_Center[0] + Move_X_Img
            New_CY = self._Zoom_Center[1] + Move_Y_Img
            Half_W = View_W / 2
            Half_H = View_H / 2
            New_CX = max(Half_W, min(self._Image_Width - Half_W, New_CX))
            New_CY = max(Half_H, min(self._Image_Height - Half_H, New_CY))
            self._Zoom_Center = (New_CX, New_CY)
            self.Load()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Drag -> {E}")
    
    def Zoom(self, Event):
        try:
            if not self._Image:
                return
            Zoom_Factor = 1.1 if Event.delta > 0 else 0.9
            New_Zoom_Scale = self._Zoom_Scale * Zoom_Factor
            New_Zoom_Scale = max(1.0, New_Zoom_Scale)
            if New_Zoom_Scale == self._Zoom_Scale:
                return
            Inner_W = max(1, self._Width - 2 * self._Border_Size)
            Inner_H = max(1, self._Height - 2 * self._Border_Size)
            if Inner_W<=0 or Inner_H<=0:
                self._Zoom_Scale = New_Zoom_Scale
                self.Load()
                return
            Mouse_X = Event.x
            Mouse_Y = Event.y
            Mouse_Rel_X = (Mouse_X - self._Border_Size) / Inner_W
            Mouse_Rel_Y = (Mouse_Y - self._Border_Size) / Inner_H
            Mouse_Rel_X = max(0.0, min(1.0, Mouse_Rel_X))
            Mouse_Rel_Y = max(0.0, min(1.0, Mouse_Rel_Y))
            View_W = self._Image_Width / self._Zoom_Scale
            View_H = self._Image_Height / self._Zoom_Scale
            View_Left = self._Zoom_Center[0] - View_W / 2
            View_Top = self._Zoom_Center[1] - View_H / 2
            Mouse_Abs_X = View_Left + Mouse_Rel_X * View_W
            Mouse_Abs_Y = View_Top + Mouse_Rel_Y * View_H
            New_View_W = self._Image_Width / New_Zoom_Scale
            New_View_H = self._Image_Height / New_Zoom_Scale
            New_View_Left = Mouse_Abs_X - Mouse_Rel_X * New_View_W
            New_View_Top = Mouse_Abs_Y - Mouse_Rel_Y * New_View_H
            New_Center_X = New_View_Left + New_View_W / 2
            New_Center_Y = New_View_Top + New_View_H / 2
            Half_W = New_View_W / 2
            Half_H = New_View_H / 2
            New_Center_X = max(Half_W, min(self._Image_Width - Half_W, New_Center_X))
            New_Center_Y = max(Half_H, min(self._Image_Height - Half_H, New_Center_Y))
            self._Zoom_Scale = New_Zoom_Scale
            self._Zoom_Center = (New_Center_X, New_Center_Y)
            self.Load()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Zoom -> {E}")
    
    def Reset(self, Event=False):
        try:
            self._Angle = 0
            self._Zoom_Scale = 1.0
            if self._Image:
                self._Zoom_Center = (self._Image_Width // 2, self._Image_Height // 2)
            self.Load()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Reset -> {E}")


import os
import time
import threading
try:
    import cv2
except Exception:
    cv2 = None
from PIL import Image as PIL_Image, ImageTk as PIL_ImageTk
from .N_GUI import GUI

PIL_Image.MAX_IMAGE_PIXELS = None

class Video:

    def __init__(self, Frame, *Args, **Kwargs):
        self._GUI = GUI._Instance
        if self._GUI is not None:
            self._Type = "Video"
            try:
                self._Frame = Frame
                self._Canvas = None
                self._Canvas_Widget = None
                self._Image_Window = None
                self._Temp_Image = None
                self._Path = False
                self._Path_Memory = False
                self._Pause = True
                self._Loop = False
                self._Rate = 1.0
                self._Queue = []
                self._Capture = None
                self._Fps = 0.0
                self._Frame_Count = 0
                self._Duration_Ms = 0
                self._Stop_Time_Ms = None
                self._Last_Frame_Bgr = None
                self._Target_Size = None
                self._Decode_Thread = None
                self._Decode_Stop = threading.Event()
                self._Decode_Wake = threading.Event()
                self._Capture_Lock = threading.RLock()
                self._Frame_Lock = threading.RLock()
                self._Pending_Frame_Bgr = None
                self._Render_Pending = False
                self._CV2_Available = cv2 is not None
                self._Bind_Window()
                self._Bind_Events()
            except Exception as E:
                self._GUI.Error(f"{self._Type} -> Init -> {E}")
        else:
            print("Error: Gluonix -> GUI Instance Has Not Been Created")

    def __str__(self):
        return "Nucleon_Glunoix_Video[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Video[]"

    def Close(self):
        try:
            self._Pause = True
            self._Decode_Stop_All()
            self._Capture_Close()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Close -> {E}")

    def __del__(self):
        try:
            self.Close()
        except:
            pass

    def Delete(self):
        try:
            if self:
                del self
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete -> {E}")

    def _After(self, Delay_Ms, Func):
        try:
            W = self._Canvas_Widget
            if W is not None and hasattr(W, "after"):
                return W.after(int(Delay_Ms), Func)
            F = self._Frame
            if F is not None and hasattr(F, "after"):
                return F.after(int(Delay_Ms), Func)
            if F is not None and hasattr(F, "_Frame") and hasattr(F._Frame, "after"):
                return F._Frame.after(int(Delay_Ms), Func)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _After -> {E}")

    def _Bind_Window(self):
        try:
            F = self._Frame
            if hasattr(F, "_Frame") and hasattr(F._Frame, "create_image") and hasattr(F._Frame, "itemconfig"):
                self._Canvas = F
                self._Canvas_Widget = F._Frame
                return
            if hasattr(F, "create_image") and hasattr(F, "itemconfig"):
                self._Canvas = None
                self._Canvas_Widget = F
                return
            self._Canvas = None
            self._Canvas_Widget = getattr(F, "_Frame", None)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Bind_Window -> {E}")

    def _Bind_Events(self):
        try:
            if hasattr(self._Frame, "Bind"):
                try:
                    self._Frame.Bind(On_Resize=self.Resize)
                except Exception:
                    pass
            W = self._Canvas_Widget
            if W is not None and hasattr(W, "bind"):
                try:
                    W.bind("<Configure>", lambda E: self.Resize(E))
                except Exception:
                    pass
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Bind_Events -> {E}")

    def Widget(self):
        try:
            return self
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Widget -> {E}")

    def _Canvas_Box(self):
        try:
            if self._Canvas is not None and hasattr(self._Canvas, "Box"):
                FX, FY, FW, FH = self._Canvas.Box()
                return float(FW), float(FH)
            W = self._Canvas_Widget
            if W is not None and hasattr(W, "winfo_width") and hasattr(W, "winfo_height"):
                return float(W.winfo_width()), float(W.winfo_height())
            return 0.0, 0.0
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Canvas_Box -> {E}")
            return 0.0, 0.0

    def _Ensure_Image_Window(self):
        try:
            W = self._Canvas_Widget
            if W is None:
                return False
            if self._Image_Window is None:
                try:
                    self._Image_Window = W.create_image(0, 0, image=None, anchor="nw")
                except Exception:
                    return False
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Ensure_Image_Window -> {E}")
            return False

    def _Compute_Draw_Rect(self, Img_W, Img_H, Canvas_W, Canvas_H):
        if Canvas_W <= 1 or Canvas_H <= 1 or Img_W <= 1 or Img_H <= 1:
            return 0, 0, 1, 1
        IA = float(Img_W) / float(Img_H)
        CA = float(Canvas_W) / float(Canvas_H)
        if IA > CA:
            DW = int(Canvas_W)
            DH = int(Canvas_W / max(1e-9, IA))
        else:
            DH = int(Canvas_H)
            DW = int(Canvas_H * IA)
        L = int((Canvas_W - DW) * 0.5)
        T = int((Canvas_H - DH) * 0.5)
        return L, T, DW, DH

    def _Render_Frame_Main(self):
        try:
            self._Render_Pending = False
            W = self._Canvas_Widget
            if W is None:
                return
            if not self._Ensure_Image_Window():
                return
            with self._Frame_Lock:
                Frame_Bgr = self._Pending_Frame_Bgr
                self._Pending_Frame_Bgr = None
            if Frame_Bgr is None:
                return
            self._Last_Frame_Bgr = Frame_Bgr
            Canvas_W, Canvas_H = self._Canvas_Box()
            Canvas_W = int(max(1, Canvas_W))
            Canvas_H = int(max(1, Canvas_H))
            H, W0 = Frame_Bgr.shape[:2]
            L, T, DW, DH = self._Compute_Draw_Rect(W0, H, Canvas_W, Canvas_H)
            Frame_Rgb = cv2.cvtColor(Frame_Bgr, cv2.COLOR_BGR2RGB) if self._CV2_Available else Frame_Bgr
            Pil = PIL_Image.fromarray(Frame_Rgb)
            if DW > 1 and DH > 1:
                Pil = Pil.resize((int(DW), int(DH)), PIL_Image.NEAREST)
            Tk = PIL_ImageTk.PhotoImage(Pil)
            self._Temp_Image = Tk
            try:
                W.itemconfig(self._Image_Window, image=Tk)
            except Exception:
                try:
                    W.itemconfigure(self._Image_Window, image=Tk)
                except Exception:
                    return
            try:
                W.coords(self._Image_Window, int(L), int(T))
            except Exception:
                pass
            try:
                W.tag_raise(self._Image_Window)
            except Exception:
                pass
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Render_Frame_Main -> {E}")

    def _Schedule_Render(self, Frame_Bgr):
        try:
            with self._Frame_Lock:
                self._Pending_Frame_Bgr = Frame_Bgr
            if self._Render_Pending:
                return
            self._Render_Pending = True
            self._After(0, self._Render_Frame_Main)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Schedule_Render -> {E}")

    def _Capture_Open(self, Source):
        try:
            if not self._CV2_Available:
                raise Exception("cv2 not found")
            Cap = cv2.VideoCapture(Source)
            if not Cap or not Cap.isOpened():
                return None
            return Cap
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Capture_Open -> {E}")
            return None

    def _Capture_Close(self):
        try:
            with self._Capture_Lock:
                if self._Capture is not None:
                    try:
                        self._Capture.release()
                    except Exception:
                        pass
                self._Capture = None
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Capture_Close -> {E}")

    def _Capture_Refresh_Info(self):
        try:
            with self._Capture_Lock:
                Cap = self._Capture
                if Cap is None:
                    self._Fps = 0.0
                    self._Frame_Count = 0
                    self._Duration_Ms = 0
                    return
                Fps = float(Cap.get(getattr(cv2, "CAP_PROP_FPS", 5)) or 0.0)
                Cnt = int(Cap.get(getattr(cv2, "CAP_PROP_FRAME_COUNT", 7)) or 0)
                if Fps <= 1e-6:
                    Fps = 30.0
                Dur = 0
                if Cnt > 0 and Fps > 1e-9:
                    Dur = int((Cnt / Fps) * 1000.0)
                self._Fps = float(Fps)
                self._Frame_Count = int(Cnt)
                self._Duration_Ms = int(Dur)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Capture_Refresh_Info -> {E}")

    def _Capture_Time_Get_Ms(self):
        try:
            with self._Capture_Lock:
                if self._Capture is None:
                    return 0
                V = self._Capture.get(getattr(cv2, "CAP_PROP_POS_MSEC", 0))
                return int(V or 0)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Capture_Time_Get_Ms -> {E}")
            return 0

    def _Capture_Time_Set_Ms(self, Value):
        try:
            with self._Capture_Lock:
                if self._Capture is None:
                    return False
                return bool(self._Capture.set(getattr(cv2, "CAP_PROP_POS_MSEC", 0), float(Value)))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Capture_Time_Set_Ms -> {E}")
            return False

    def _Capture_Pos_Get(self):
        try:
            with self._Capture_Lock:
                if self._Capture is None:
                    return 0.0
                Cnt = int(self._Frame_Count or 0)
                if Cnt <= 0:
                    F = float(self._Capture.get(getattr(cv2, "CAP_PROP_POS_MSEC", 0)) or 0.0)
                    D = float(self._Duration_Ms or 0.0)
                    return float(F / max(1e-9, D)) if D > 0 else 0.0
                Cur = float(self._Capture.get(getattr(cv2, "CAP_PROP_POS_FRAMES", 1)) or 0.0)
                return float(Cur / max(1.0, float(Cnt)))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Capture_Pos_Get -> {E}")
            return 0.0

    def _Capture_Pos_Set(self, Value):
        try:
            V = float(Value)
            V = max(0.0, min(1.0, V))
            with self._Capture_Lock:
                if self._Capture is None:
                    return False
                Cnt = int(self._Frame_Count or 0)
                if Cnt > 0:
                    Target = int(V * float(Cnt))
                    return bool(self._Capture.set(getattr(cv2, "CAP_PROP_POS_FRAMES", 1), float(Target)))
                D = int(self._Duration_Ms or 0)
                if D > 0:
                    return bool(self._Capture.set(getattr(cv2, "CAP_PROP_POS_MSEC", 0), float(V * D)))
                return False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Capture_Pos_Set -> {E}")
            return False

    def _Decode_Loop(self):
        try:
            Last_T = time.time()
            while not self._Decode_Stop.is_set():
                if self._Pause:
                    self._Decode_Wake.wait(0.05)
                    continue
                with self._Capture_Lock:
                    Cap = self._Capture
                if Cap is None:
                    time.sleep(0.02)
                    continue
                Fps = float(self._Fps or 30.0)
                Rate = float(self._Rate or 1.0)
                Step_S = (1.0 / max(1e-9, Fps)) / max(1e-9, Rate)
                Now = time.time()
                Dt = Now - Last_T
                if Dt < Step_S:
                    time.sleep(min(0.01, Step_S - Dt))
                    continue
                Last_T = Now
                Ok = False
                Frame = None
                with self._Capture_Lock:
                    if self._Capture is None:
                        continue
                    Ok, Frame = self._Capture.read()
                if not Ok or Frame is None:
                    if self._Loop:
                        self.Position_Set(0.0)
                        continue
                    self._Pause = True
                    self.Play_Next()
                    continue
                if self._Stop_Time_Ms is not None:
                    Cur_Ms = self._Capture_Time_Get_Ms()
                    if Cur_Ms >= int(self._Stop_Time_Ms):
                        if self._Loop:
                            self.Time_Set(0)
                            continue
                        self._Pause = True
                        self.Play_Next()
                        continue
                self._Schedule_Render(Frame)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Decode_Loop -> {E}")

    def _Decode_Start(self):
        try:
            if self._Decode_Thread is not None and self._Decode_Thread.is_alive():
                return
            self._Decode_Stop.clear()
            self._Decode_Wake.set()
            self._Decode_Thread = threading.Thread(target=self._Decode_Loop, daemon=True)
            self._Decode_Thread.start()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Decode_Start -> {E}")

    def _Decode_Stop_All(self):
        try:
            self._Decode_Stop.set()
            self._Decode_Wake.set()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Decode_Stop_All -> {E}")

    def Set(self, Path=''):
        try:
            if not self._CV2_Available:
                raise Exception("cv2 not found")
            if Path:
                self._Path = Path
                if self._Path != self._Path_Memory:
                    self._Path_Memory = self._Path
                    self.Open()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Set -> {E}")

    def Open(self):
        try:
            if not self._CV2_Available:
                raise Exception("cv2 not found")
            if not self._Path:
                return
            self.Stop()
            Src = self._Path
            if isinstance(Src, str):
                if Src.isdigit():
                    Src = int(Src)
            Cap = self._Capture_Open(Src)
            if Cap is None:
                self._GUI.Error(f"{self._Type} -> Open -> Could not open: {self._Path}")
                return
            with self._Capture_Lock:
                self._Capture = Cap
            self._Stop_Time_Ms = None
            self._Capture_Refresh_Info()
            self._Decode_Start()
            self._Prime_Frame()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Open -> {E}")

    def Open_With(self, Path, Start=0, Stop=None, Cache_Ms=300, Hw=True):
        try:
            if not self._CV2_Available:
                raise Exception("cv2 not found")
            if not Path:
                return
            self._Path = Path
            self._Path_Memory = Path
            self.Stop()
            Cap = self._Capture_Open(Path)
            if Cap is None:
                self._GUI.Error(f"{self._Type} -> Open_With -> Could not open: {Path}")
                return
            with self._Capture_Lock:
                self._Capture = Cap
            self._Capture_Refresh_Info()
            if Start:
                self.Time_Set(int(float(Start) * 1000.0))
            if Stop is not None:
                self._Stop_Time_Ms = int(float(Stop) * 1000.0)
            else:
                self._Stop_Time_Ms = None
            self._Decode_Start()
            self._Prime_Frame()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Open_With -> {E}")

    def _Prime_Frame(self):
        try:
            with self._Capture_Lock:
                Cap = self._Capture
                if Cap is None:
                    return
                Ok, Frame = Cap.read()
                if Ok and Frame is not None:
                    self._Last_Frame_Bgr = Frame
                    self._Schedule_Render(Frame)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Prime_Frame -> {E}")

    def Playing(self):
        try:
            return bool((not self._Pause) and (self._Capture is not None))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Playing -> {E}")
            return False

    def State(self):
        try:
            if self._Capture is None:
                return "stopped"
            if self._Pause:
                return "paused"
            return "playing"
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> State -> {E}")
            return "unknown"

    def Stop(self):
        try:
            self._Pause = True
            self._Stop_Time_Ms = None
            self._Capture_Close()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Stop -> {E}")

    def Pause(self):
        try:
            self._Pause = True
            self._Decode_Wake.set()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Pause -> {E}")

    def Play(self):
        try:
            if self._Capture is None and self._Path:
                self.Open()
            if self._Capture is None:
                return
            self._Pause = False
            self._Decode_Wake.set()
            self._Decode_Start()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Play -> {E}")

    def Play_BlockingStart(self, Timeout=2.0):
        try:
            T0 = time.time()
            self.Play()
            while time.time() - T0 < float(Timeout):
                if self._Capture is not None and not self._Pause:
                    return True
                time.sleep(0.02)
            return False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Play_BlockingStart -> {E}")
            return False

    def Length(self):
        try:
            return int(self._Duration_Ms or 0)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Length -> {E}")
            return 0

    def Time_Get(self):
        try:
            return int(self._Capture_Time_Get_Ms())
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Time_Get -> {E}")
            return 0

    def Time_Set(self, Value):
        try:
            Ok = self._Capture_Time_Set_Ms(int(Value))
            if Ok:
                self._Prime_Frame()
            return Ok
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Time_Set -> {E}")
            return False

    def Position_Get(self):
        try:
            return float(self._Capture_Pos_Get())
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Position_Get -> {E}")
            return 0.0

    def Position_Set(self, Value):
        try:
            Ok = self._Capture_Pos_Set(float(Value))
            if Ok:
                self._Prime_Frame()
            return Ok
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Position_Set -> {E}")
            return False

    def Frame_Step(self):
        try:
            if self._Capture is None:
                return False
            Was_Pause = bool(self._Pause)
            self._Pause = True
            with self._Capture_Lock:
                Cap = self._Capture
                if Cap is None:
                    return False
                Ok, Frame = Cap.read()
            if Ok and Frame is not None:
                self._Last_Frame_Bgr = Frame
                self._Schedule_Render(Frame)
                self._Pause = Was_Pause
                return True
            self._Pause = Was_Pause
            return False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Frame_Step -> {E}")
            return False

    def Rate_Set(self, Value):
        try:
            self._Rate = float(Value)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Rate_Set -> {E}")
            return False

    def Rate_Get(self):
        try:
            return float(self._Rate or 1.0)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Rate_Get -> {E}")
            return 1.0

    def Loop(self, Enabled):
        try:
            self._Loop = bool(Enabled)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Loop -> {E}")
            return False

    def Snapshot(self, Path, Width=0, Height=0):
        try:
            Frame = self._Last_Frame_Bgr
            if Frame is None:
                return False
            Out = Frame
            if int(Width) > 0 and int(Height) > 0 and self._CV2_Available:
                Out = cv2.resize(Out, (int(Width), int(Height)))
            Ok = False
            if self._CV2_Available:
                Ok = bool(cv2.imwrite(str(Path), Out))
            return Ok
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Snapshot -> {E}")
            return False

    def Queue(self, Paths):
        try:
            self._Queue = list(Paths) if Paths else []
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Queue -> {E}")

    def Play_Next(self):
        try:
            if self._Queue:
                Nxt = self._Queue.pop(0)
                self.Open_With(Nxt)
                self.Play()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Play_Next -> {E}")

    def Resize(self, Event=False):
        try:
            if self._Last_Frame_Bgr is None:
                return
            self._After(0, self._Render_Frame_Main)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Resize -> {E}")


# IMPORT LIBRARIES
import os
import math
from PIL import Image as PIL_Image, ImageTk as PIL_ImageTk
import tkinter as TK
import tkinter.font as TK_Font
from .N_GUI import GUI
from .N_Frame import Frame
from .N_Custom import Event_Bind

class Tree:

    def __init__(self, Main, *args, **kwargs):
        self._GUI = GUI._Instance
        if self._GUI is not None:
            self._Type = "Tree"
            try:
                self._Config = ['Name', 'Auto_Dark', 'Background', 'Light_Background', 'Dark_Background', 'Foreground', 'Light_Foreground', 'Dark_Foreground', 'Border_Color', 'Light_Border_Color', 'Dark_Border_Color', 'Border_Size', 'Resize', 'Popup', 'Display', 'Left', 'Top', 'Width', 'Height', 'Font_Size', 'Font_Weight', 'Scrollbar_Size', 'Scrollbar_Active', 'Background_Selected', 'Foreground_Selected', 'Hover_Background', 'Light_Hover_Background', 'Dark_Hover_Background', 'Hover_Foreground', 'Light_Hover_Foreground', 'Dark_Hover_Foreground', 'Hover_Border_Color', 'Light_Hover_Border_Color', 'Dark_Hover_Border_Color']
                self._Initialized = False
                self._Name = False
                self._Last_Name = False
                self._Tag = []
                self._Item = []
                self._Image = {}
                self._Resize = True
                self._Popup = False
                self._Display = True
                self._Size_Update = False
                self._Main = Main
                self._Style_Name = "CustomStyle_" + str(id(self)) + ".Treeview"
                self._GUI._Style.layout(self._Style_Name, [])
                self._Frame = Frame(self._Main)
                self._Widget = TK.ttk.Treeview(self._Frame._Frame, show="tree")
                self._Widget.configure(style=self._Style_Name)
                self._Scrollbar_Vertical = TK.Scrollbar(self._Frame._Frame)
                self._Scrollbar_Horizontal = TK.Scrollbar(self._Frame._Frame)
                self._Scrollbar_Size = 20
                self._Scrollbar_Active = False
                self._Border_Color = '#000000'
                self._Border_Size = 0
                self._Background = self._Main._Background
                self._Background_Main = True
                self._Foreground = '#000000'
                self._Hover_Background = False
                self._Hover_Foreground = False
                self._Hover_Border_Color = False
                self._Last_Background = False
                self._Last_Foreground = False
                self._Last_Border_Color = False
                self._Background_Selected = "#d0e4f5"
                self._Foreground_Selected = "#000000"
                self._Font_Size = 12
                self._Font_Weight = 'normal'
                self._Auto_Dark = True
                self._Width_Item = None
                self._Width_Text = ''
                self._On_Show = False
                self._On_Hide = False
            except Exception as E:
                self._GUI.Error(f"{self._Type} -> Init -> {E}")
        else:
            print("Error: Gluonix -> GUI Instance Has Not Been Created")

    def __str__(self):
        return "Nucleon_Glunoix_Tree[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Tree[]"
    
    def Copy(self, Name=False, Main=False):
        try:
            if not Main:
                Main = self._Main
            Instance = type(self)(Main)
            for Key in self._Config:
                if hasattr(self, "_"+Key):
                    setattr(Instance, "_"+Key, getattr(self, "_"+Key))
            if Name:
                setattr(Instance, "_Name", Name)
            Instance.Create()
            return Instance
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Copy -> {E}")
        
    def Delete(self):
        try:
            self._Main._Widget.remove(self)
            self._Widget.destroy()
            self._Scrollbar_Vertical.destroy()
            self._Scrollbar_Horizontal.destroy()
            self._Frame.Delete()
            if self:
                del self
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete -> {E}")
            
    def Hide(self):
        try:
            self._Frame.Hide()
            self._Display = False
            if self._On_Hide:
                self._On_Hide()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Hide -> {E}")
            
    def Show(self):
        try:
            self._Frame.Show()
            if self._Scrollbar_Active:
                self._Scrollbar_Vertical.place(relx=(self._Width-self._Scrollbar_Size)/self._Width, rely=0, relwidth=self._Scrollbar_Size/self._Width, relheight=1)
                self._Scrollbar_Horizontal.place(relx=0, rely=(self._Height-self._Scrollbar_Size)/self._Height, relwidth=1, relheight=self._Scrollbar_Size/self._Height)
                self._Widget.place(relx=0, rely=0, relwidth=(self._Width-self._Scrollbar_Size)/self._Width, relheight=(self._Height-self._Scrollbar_Size)/self._Height)
            else:
                self._Scrollbar_Vertical.place_forget()
                self._Scrollbar_Horizontal.place_forget()
                self._Widget.place(relx=0, rely=0, relwidth=1.0, relheight=1.0)
            self._Display = True
            if self._On_Show:
                self._On_Show()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Show -> {E}")
    
    def Grab(self, Path=False):
        try:
            return self._GUI.Grab_Widget(Path=Path, Widget=self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Grab -> {E}")
            
    def Animate(self, Hide=False, Thread=True):
        try:
            self._Frame.Animate(Widget=self._Widget, Thread=Thread)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate -> {E}")
            self.Animate_Cancel()
            
    def Animate_Cancel(self):
        try:
            self._Frame.Animate_Cancel()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate_Cancel -> {E}")
            
    def Scroll(self, E):
        try:
            if E.delta:
                self._Widget.yview_scroll(int(-1*(E.delta/120)), "units")
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Scroll -> {E}")

    def Export(self, Path="TreeStructure.txt"):
        try:
            if os.path.exists(Path):
                os.remove(Path)
            with open(Path, "w") as File:
                self.Export_All(File)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Export -> {E}")

    def Export_All(self, File, Parent="", Indent=0):
        try:
            Line = " " * Indent + self._Widget.item(Parent, "text") + "\n"
            File.write(Line)
            for Child in self._Widget.get_children(Parent):
                self.Export_All(File, Child, Indent + 4)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Export_All -> {E}")
            
    def Get(self, ID=False):
        try:
            if not ID:
                ID = self._Widget.focus()
            return self._Widget.item(ID)['values']
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Get -> {E}")
            
    def Get_All(self, ID=False):
        try:
            if not ID:
                ID = self._Widget.focus()
            return self._Widget.item(ID)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Get_All -> {E}")
            
    def Get_Selected(self):
        try:
            Return = []
            for Each in self._Widget.selection():
                Return.append(self._Widget.item(Each))
            return Return
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Get_Selected -> {E}")
            
    def Remove(self, ID):
        try:
            self._Widget.delete(ID)
            self._Item.remove(ID)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Remove -> {E}")
            
    def Remove_All(self):
        try:
            for Each in self._Item:
                if self._Widget.exists(Each):
                    self.Remove(Each)
            self._Item = []
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Remove_All -> {E}")
            
    def Remove_Selected(self):
        try:
            for Each in self._Widget.selection():
                self.Remove(Each)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Remove_Selected -> {E}")
            
    def Edit(self, Name=False, Value=False, Tag=False, ID=False):
        try:
            if not ID:
                ID = self._Widget.focus()
            if Name:
                self._Widget.item(ID, text=Name)
                self.Update_Tree(Name, ID)
            if Value:
                self._Widget.item(ID, value=Value)
            if Tag:
                self._Widget.item(ID, tags=(Tag))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Edit -> {E}")
            
    def Add(self, Name, Parent='', Index='end', Tag='Default', Value=[], ID=None, Path=None):
        try:
            if Index!='end':
                if not isinstance(Index, int):
                    Index = int(self.Index(Index))
            Insert_Args = {'parent': Parent, 'index': Index, 'text': f' {Name}', 'values': Value, 'tags': Tag}
            if ID is not None:
                Insert_Args['iid'] = ID
            if Path is not None and os.path.exists(Path):
                if Path not in self._Image:
                    Image_Pil = PIL_Image.open(Path)
                    Height = math.ceil(self._Font_Size * 1.5)
                    Aspect_Ratio = Image_Pil.width / Image_Pil.height
                    Width = int(Height * Aspect_Ratio)
                    Image_Pil = Image_Pil.resize((Width, Height), PIL_Image.NEAREST)
                    Image = PIL_ImageTk.PhotoImage(Image_Pil)
                    self._Image[Path] = Image
                else:
                    Image = self._Image[Path]
                Insert_Args['image'] = Image
            Item = self._Widget.insert(**Insert_Args)
            self._Item.append(Item)
            self.Update_Tree(Name, Item)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Add -> {E}")
            
    def Update_Tree(self, Name, Item):
        try:
            Font_Temp = TK_Font.Font(family='Times New Roman', size=self._Font_Size, weight=self._Font_Weight)
            Level_Item = self.Level(Item)
            Indent_Pixels = Level_Item * int(self._Font_Size*4)
            Text_Column = f" {Name}"
            Width_Text = Font_Temp.measure(Text_Column)
            Width_New = Indent_Pixels + Width_Text
            Width_Current = self._Widget.column("#0", "width")
            Width_Current = int(Width_Current) if Width_Current else 0
            if Width_New > Width_Current:
                self._Widget.column("#0", width=Width_New, stretch=False)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Update_Tree_Column_Width_For_Row -> {E}")

    def Index(self, Item):
        try:
            return self._Widget.index(Item)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Index -> {E}")
            
    def Level(self, Item):
        try:
            Level = 0
            Parent = self._Widget.parent(Item)
            while Parent:
                Level += 1
                Parent = self._Widget.parent(Parent)
            return Level
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Level -> {E}")
            
    def Tag(self, Name, Foreground=False, Background=False):
        try:
            if not Foreground:
                Foreground = self._Foreground
            if not Background:
                Background = self._Background
            if Name not in self._Tag:
                self._Widget.tag_configure(Name, font=('Times New Roman', self._Font_Size, self._Font_Weight), foreground=Foreground, background=Background)
                self._Tag.append(Name)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Tag -> {E}")
            
    def Current(self):
        try:
            return self._Widget.focus()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Current -> {E}")
            
    def Selected(self):
        try:
            Temp_List = []
            for Each in self._Widget.selection():
                Temp_List.append(Each)
            return Temp_List
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Selected -> {E}")
            
    def Child(self, ID=False):
        try:
            if ID:
                return self._Widget.get_children(ID)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Child -> {E}")
            
    def Parent(self, ID=False):
        try:
            if ID:
                return self._Widget.parent(ID)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Parent -> {E}")
            
    def Expand(self, ID=False):
        try:
            if not ID:
                ID = self._Widget.focus()
            self._Widget.item(ID, open=True)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Expand -> {E}")
            
    def Select(self, ID=None, Keep=False):
        try:
            if ID:
                if Keep:
                    self._Widget.selection_add(ID)
                else:
                    self._Widget.selection_set(ID)
                if isinstance(ID, (list, tuple)):
                    if len(ID) > 0:
                        self._Widget.focus(ID[0])
                        self._Widget.see(ID[0]) 
                else:
                    self._Widget.focus(ID)
                    self._Widget.see(ID)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Select -> {E}")
            
    def Unselect(self, ID=None):
        try:
            if ID:
                self._Widget.selection_remove(ID)
            else:
                self._Widget.selection_remove(self._Widget.selection())
            self._Widget.focus('')
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Unselect -> {E}")

    def Widget(self):
        try:
            return self._Widget
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Widget -> {E}")
            
    def Bind(self, **Input):
        try:
            if 'On_Show' in Input:
                self._On_Show = Input['On_Show']
            if 'On_Hide' in Input:
                self._On_Hide = Input['On_Hide']
            Event_Bind(self._Widget, **Input)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind -> {E}")
            
    def On_Hover_In(self, E):
        try:
            Config = {}
            if self._Hover_Background:
                self._Last_Background = self._Background
                Config['Background'] = self._Hover_Background
            if self._Hover_Foreground:
                self._Last_Foreground = self._Foreground
                Config['Foreground'] = self._Hover_Foreground
            if self._Hover_Border_Color:
                self._Last_Border_Color = self._Border_Color
                Config['Border_Color'] = self._Hover_Border_Color
            if len(Config)>0:
                self.Config(**Config)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Hover_In -> {E}")
            
    def On_Hover_Out(self, E):
        try:
            Config = {}
            if self._Hover_Background and self._Last_Background:
                Config['Background'] = self._Last_Background if self._Background==self._Hover_Background else self._Background
            if self._Hover_Foreground and self._Last_Foreground:
                Config['Foreground'] = self._Last_Foreground if self._Foreground==self._Hover_Foreground else self._Foreground
            if self._Hover_Border_Color and self._Last_Border_Color:
                Config['Border_Color'] = self._Last_Border_Color if self._Border_Color==self._Hover_Border_Color else self._Border_Color
            if len(Config)>0:
                self.Config(**Config)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Hover_Out -> {E}")
            
    def Config_Get(self, *Input):
        try:
            Return = {}
            for Each in self._Config:
                if Each in Input:
                    Return[Each] = getattr(self, "_"+Each)
            return Return
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config_Get -> {E}")
                
    def Config(self, **Input):
        try:
            Run = False
            for Each in self._Config:
                if Each in Input:
                    Value = Input[Each]
                    setattr(self, "_"+Each, Value)
                    Run = True
            self._Frame.Config(**Input)
            if "Width" in Input or "Height" in Input or "Left" in Input or "Top" in Input:
                self._Size_Update = True
            if self._Initialized and Run:
                self.Create()
            if "Background" in Input:
                self._Background_Main = not bool(Input["Background"])
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config -> {E}")
            
    def Move(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left += Left
            if Top is not None:
                self._Top += Top
            if Left is not None or Top is not None:
                self.Position(Left=self._Left, Top=self._Top)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Move -> {E}")
            
    def Center(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left-self._Width/2
            if Top is not None:
                self._Top = Top-self._Height/2
            if Left is not None or Top is not None:
                self.Position(Left=self._Left, Top=self._Top)
            return [self._Left+self._Width/2, self._Top+self._Height/2]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Center -> {E}")
            
    def Position(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left
            if Top is not None:
                self._Top = Top
            if Left is not None or Top is not None:
                self._Frame.Position(Left=self._Left, Top=self._Top)
            return self._Frame.Position()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Position -> {E}")
            
    def Size(self, Width=False, Height=False):
        try:
            if Width:
                self._Width = Width
            if Height:
                self._Height = Height
            if Width or Height:
                self._Frame.Size(Width=self._Width, Height=self._Height)
                self.Create()
            return self._Frame.Size()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Size -> {E}")
    
    def Enlarge(self, Value=None):
        try:
            if Value is not None:
                self._Left -= Value
                self._Top -= Value
                self._Width += Value
                self._Height += Value
                self._Frame.Enlarge(Value)
                self.Create()
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Enlarge -> {E}")
    
    def Shrink(self, Value=None):
        try:
            if Value is not None:
                self._Left += Value
                self._Top += Value
                self._Width -= Value
                self._Height -= Value
                self._Frame.Shrink(Value)
                self.Create()
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Shrink -> {E}")
            
    def Box(self):
        try:
            return self._Frame.Box()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Box -> {E}")
        
    def Ratio(self):
        try:
            return self._Frame.Ratio()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Ratio -> {E}")
        
    def Locate(self, Width, Height, Left, Top):
        try:
            Width = self._Width*(Width/100)
            Height = self._Height*(Height/100)
            Left = self._Width*(Left/100)-self._Border_Size
            Top = self._Height*(Top/100)-self._Border_Size
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate -> {E}")
        
    def Locate_Reverse(self, Width, Height, Left, Top):
        try:
            Width = round((Width/self._Width)*100, 3)
            Height = round((Height/self._Height)*100, 3)
            Left =  round((Left/self._Width)*100, 3)
            Top =  round((Top/self._Height)*100, 3)
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate_Reverse -> {E}")
            
    def Create(self):
        try:
            if not self._Background:
                self._Background = self._Main._Background
                if not hasattr(self, "_Light_Background"):
                    setattr(self, "_Light_Background", self._Background)
                if not hasattr(self, "_Dark_Background"):
                    setattr(self, "_Dark_Background", self._GUI.Invert(self._Background))
            if self._Auto_Dark and not self._GUI._Dark_Mode:
                self.Update_Color()
            if not self._Initialized:
                self.Update_Color()
                self._Frame.Config(Width=self._Width, Height=self._Height, Left=self._Left, Top=self._Top)
                self._Frame.Config(Background=self._Background, Border_Size=self._Border_Size, Border_Color=self._Border_Color)
                self._Frame.Create()
                Event_Bind(self._Widget, On_Hover_In=lambda E: self.On_Hover_In(E), On_Hover_Out=lambda E: self.On_Hover_Out(E))
                self._Widget.bind("<MouseWheel>", self.Scroll)
                self.Tag(Name='Default', Foreground=self._Foreground)
                if not self._Display:
                    self.Hide()
                self._Main._Widget.append(self)
                self._Initialized = True
            self._Widget.tag_configure('Default', font=('Times New Roman', self._Font_Size, self._Font_Weight), foreground=self._Foreground, background=self._Background)
            self._GUI._Style.configure(self._Style_Name, rowheight=self._Font_Size*2, background=self._Background, highlightthickness=0, bd=0, borderwidth=0)
            self._GUI._Style.map(self._Style_Name, background=[("selected", self._Background_Selected)], foreground=[("selected", self._Foreground_Selected)])
            self._Widget.configure(style=self._Style_Name)
            self._Scrollbar_Vertical.config(orient="vertical", command=self._Widget.yview, width=self._Scrollbar_Size)
            self._Scrollbar_Horizontal.config(orient="horizontal", command=self._Widget.xview, width=self._Scrollbar_Size)
            self._Widget.configure(xscrollcommand=self._Scrollbar_Horizontal.set, yscrollcommand=self._Scrollbar_Vertical.set)
            if self._Display:
                self.Show()
            else:
                self.Hide()
            if self._Name!=self._Last_Name:
                if self._Last_Name:
                    if self._Last_Name in self._Main.__dict__:
                        del self._Main.__dict__[self._Last_Name]
                if self._Name:
                    self._Main.__dict__[self._Name] = self
                self._Last_Name = self._Name
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Create -> {E}")
            
    def Update_Color(self):
        try:
            self._GUI.Initiate_Colors(self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Update_Color -> {E}")


# IMPORT LIBRARIES
import os
import time
try:
    import vlc
except Exception:
    vlc = None
from .N_GUI import GUI

class Player:

    def __init__(self, Frame, *Args, **Kwargs):
        self._GUI = GUI._Instance
        if self._GUI is not None:
            self._Type = "Player"
            try:
                self._Frame = Frame
                self._VLC_Args = ['--quiet', '--no-xlib']
                self._VLC_Available = vlc is not None
                self._VLC = vlc.Instance(self._VLC_Args) if self._VLC_Available else None
                self._Player = self._VLC.media_player_new() if self._VLC_Available else None
                if self._Player:
                    self.Bind_Window()
                    self.Bind_Events()
                self._Media = False
                self._Pause = True
                self._Path = False
                self._Path_Memory = False
                self._Subtitle = False
                self._Queue = []
                self._EQ = None
            except Exception as E:
                self._GUI.Error(f"{self._Type} -> Init -> {E}")
        else:
            print("Error: Gluonix -> GUI Instance Has Not Been Created")

    def __str__(self):
        return "Nucleon_Glunoix_Player[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Player[]"
            
    def __del__(self):
        try:
            self.Close()
        except:
            pass
            
    def Close(self):
        try:
            self.Stop()
            if self._Player:
                try:
                    self._Player.release()
                except Exception:
                    pass
            if self._Media:
                try:
                    self._Media.release()
                except Exception:
                    pass
            if self._VLC:
                try:
                    self._VLC.release()
                except Exception:
                    pass
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Close -> {E}")
        
    def Delete(self):
        try:
            if self:
                del self
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete -> {E}")
            
    def Set(self, Path=''):
        try:
            if not self._VLC_Available:
                raise Exception("python-vlc not found")
            if Path:
                self._Path = Path
                if self._Path!=self._Path_Memory:
                    self._Path_Memory = self._Path
                    self.Open()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Set -> {E}")
            
    def Widget(self):
        try:
            return self._Player
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Widget -> {E}")
            
    def Open(self):
        try:
            if not self._VLC_Available:
                raise Exception("python-vlc not found")
            if self._Path:
                self.Stop()
                if self._Subtitle:
                    self.Subtitle_Off()
                    self._Subtitle = False
                self._Media = self._VLC.media_new(self._Path)
                self._Player.set_media(self._Media)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Open -> {E}")
            
    def Open_With(self, Path, Start=0, Stop=None, Cache_Ms=300, Hw=True):
        try:
            if not self._VLC_Available:
                raise Exception("python-vlc not found")
            if Path:
                self.Stop()
                if self._Subtitle:
                    self.Subtitle_Off()
                    self._Subtitle = False
                M = self._VLC.media_new(Path)
                M.add_option(f":network-caching={int(Cache_Ms)}")
                if Start:
                    M.add_option(f":start-time={int(Start)}")
                if Stop is not None:
                    M.add_option(f":stop-time={int(Stop)}")
                if Hw:
                    M.add_option(":hwdec=auto")
                self._Media = M
                self._Player.set_media(M)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Open_With -> {E}")
            
    def Playing(self):
        try:
            return bool(self._Player.is_playing()) if self._Player else False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> State -> {E}")
            
    def State(self):
        try:
            return self._Player.get_state()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> State -> {E}")
            
    def Stop(self):
        try:
            if self._Player:
                self._Player.stop()
                self._Pause = True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Stop -> {E}")
            
    def Pause(self):
        try:
            if self._Player:
                self._Player.pause()
                self._Pause = True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Pause -> {E}")
            
    def Play(self):
        try:
            if self._Player:
                self._Player.play()
                self._Pause = False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Play -> {E}")
            
    def Play_BlockingStart(self, Timeout=2.0):
        try:
            T0 = time.time()
            self.Play()
            while time.time() - T0 < float(Timeout):
                if self._Player and self._Player.is_playing():
                    return True
                time.sleep(0.02)
            return False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Play_BlockingStart -> {E}")
            
    def Length(self):
        try:
            return self._Player.get_length()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Length -> {E}")
            
    def Time_Get(self):
        try:
            return self._Player.get_time()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Time_Get -> {E}")
            
    def Time_Set(self, Value):
        try:
            return self._Player.set_time(Value)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Time_Set -> {E}")
            
    def Position_Get(self):
        try:
            return self._Player.get_position()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Position_Get -> {E}")
            
    def Position_Set(self, Value):
        try:
            return self._Player.set_position(float(Value))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Position_Set -> {E}")
            
    def Frame_Step(self):
        try:
            return self._Player.next_frame()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Frame_Step -> {E}")
            
    def Rate_Set(self, Value):
        try:
            return self._Player.set_rate(float(Value))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Rate_Set -> {E}")
            
    def Rate_Get(self):
        try:
            return self._Player.get_rate()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Rate_Get -> {E}")
            
    def Loop(self, Enabled):
        try:
            return self._Player.set_repeat(bool(Enabled))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Loop -> {E}")
            
    def Audio_Get(self):
        try:
            return self._Player.audio_get_track_description()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Get -> {E}")
            
    def Audio_Set(self, ID):
        try:
            return self._Player.audio_set_track(ID)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Set -> {E}")
            
    def Audio_Tracks(self):
        try:
            Desc = self._Player.audio_get_track_description()
            return [(D.id, D.name) for D in Desc] if Desc else []
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Tracks -> {E}")
            
    def Audio_Set_ByName(self, Name):
        try:
            Tracks = self.Audio_Tracks()
            for Tid, N in Tracks:
                if N and str(N).lower() == str(Name).lower():
                    return self._Player.audio_set_track(Tid)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Set_ByName -> {E}")
            
    def Audio_Channel_Set(self, Channel):
        try:
            return self._Player.audio_set_channel(int(Channel))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Channel_Set -> {E}")
            
    def Audio_Delay_Set(self, Microseconds):
        try:
            return self._Player.audio_set_delay(int(Microseconds))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Delay_Set -> {E}")
            
    def Audio_Delay_Get(self):
        try:
            return self._Player.audio_get_delay()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Delay_Get -> {E}")
            
    def Audio_Equalizer_Enable(self, Bands=None, Preamp=0.0):
        try:
            Eq = vlc.AudioEqualizer() if self._VLC_Available else None
            if Eq is None:
                return
            if Bands:
                for I, Gain in enumerate(Bands):
                    Eq.set_amp_at_index(float(Gain), I)
            Eq.set_preamp(float(Preamp))
            self._Player.set_equalizer(Eq)
            self._EQ = Eq
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Equalizer_Enable -> {E}")
            
    def Audio_Devices(self):
        try:
            return self._Player.audio_output_device_get()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Audio_Devices -> {E}")
            
    def Mute(self, Value=None):
        try:
            if isinstance(Value, bool) or Value in (0, 1):
                self._Player.audio_set_mute(bool(Value))
            return self._Player.audio_get_mute()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Mute -> {E}")
            
    def Volume_Get(self):
        try:
            return self._Player.audio_get_volume()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Volume_Get -> {E}")
            
    def Volume_Set(self, Value):
        try:
            return self._Player.audio_set_volume(Value)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Volume_Set -> {E}")
            
    def Volume_Fade(self, Target, Duration=0.5, Steps=10):
        try:
            Start = self._Player.audio_get_volume()
            Steps_I = int(Steps)
            Dur = float(Duration)
            for I in range(1, Steps_I+1):
                V = Start + (int(Target) - Start)*I/Steps_I
                self._Player.audio_set_volume(int(V))
                time.sleep(Dur/Steps_I)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Volume_Fade -> {E}")
            
    def Subtitle(self, Path):
        try:
            if self._Subtitle:
                self.Subtitle_Off()
                self._Subtitle = False
            self._Player.add_slave(vlc.MediaSlaveType.subtitle, Path, True)
            self._Subtitle = True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Subtitle -> {E}")
            
    def Subtitle_On(self):
        try:
            if self._Subtitle:
                self._Player.video_set_spu(0)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Subtitle_On -> {E}")
            
    def Subtitle_Off(self):
        try:
            if self._Subtitle:
                self._Player.video_set_spu(-1)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Subtitle_Off -> {E}")
            
    def Subtitle_Tracks(self):
        try:
            Desc = self._Player.video_get_spu_description()
            return [(D.id, D.name) for D in Desc] if Desc else []
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Subtitle_Tracks -> {E}")
            
    def Subtitle_Set(self, ID):
        try:
            return self._Player.video_set_spu(int(ID))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Subtitle_Set -> {E}")
            
    def Subtitle_Delay_Set(self, Microseconds):
        try:
            return self._Player.video_set_spu_delay(int(Microseconds))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Subtitle_Delay_Set -> {E}")
            
    def Subtitle_Delay_Get(self):
        try:
            return self._Player.video_get_spu_delay()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Subtitle_Delay_Get -> {E}")
            
    def Aspect_Set(self, Ratio):
        try:
            if Ratio is None:
                return self._Player.video_set_aspect_ratio(None)
            return self._Player.video_set_aspect_ratio(str(Ratio).encode())
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Aspect_Set -> {E}")
            
    def Scale_Set(self, Factor):
        try:
            if Factor is None:
                return self._Player.video_set_scale(-1)
            return self._Player.video_set_scale(float(Factor))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Scale_Set -> {E}")
            
    def Snapshot(self, Path, Width=0, Height=0):
        try:
            return self._Player.video_take_snapshot(0, Path, int(Width), int(Height))
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Snapshot -> {E}")
            
    def Queue(self, Paths):
        try:
            self._Queue = list(Paths) if Paths else []
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Queue -> {E}")
            
    def Next(self):
        try:
            if self._Queue:
                Nxt = self._Queue.pop(0)
                self.Open_With(Nxt)
                self.Play()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Next -> {E}")
            
    def Bind_Events(self):
        try:
            if not self._Player:
                return
            Em = self._Player.event_manager()
            Em.event_attach(vlc.EventType.MediaPlayerEndReached, self.On_End)
            Em.event_attach(vlc.EventType.MediaPlayerEncounteredError, self.On_Err)
            Em.event_attach(vlc.EventType.MediaPlayerPlaying, self.On_Play)
            Em.event_attach(vlc.EventType.MediaPlayerPaused, self.On_Pause)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind_Events -> {E}")
            
    def On_End(self, Event):
        try:
            self.Next()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_End -> {E}")
            
    def On_Err(self, Event):
        try:
            self._GUI.Error(f"{self._Type} -> VLC Error")
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Err -> {E}")
            
    def On_Play(self, Event):
        try:
            self._Pause = False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Play -> {E}")
            
    def On_Pause(self, Event):
        try:
            self._Pause = True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Pause -> {E}")
            
    def Bind_Window(self):
        try:
            if hasattr(self._Frame, "_Frame") and hasattr(self._Frame._Frame, "winfo_id") and self._Player:
                Wid = self._Frame._Frame.winfo_id()
                try:
                    self._Player.set_hwnd(Wid)
                except Exception:
                    try:
                        self._Player.set_xwindow(Wid)
                    except Exception:
                        try:
                            self._Player.set_nsobject(int(Wid))
                        except Exception:
                            pass
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind_Window -> {E}")
            
    def Close(self):
        try:
            self.Stop()
            if self._Player:
                try:
                    self._Player.release()
                except Exception:
                    pass
            if self._Media:
                try:
                    self._Media.release()
                except Exception:
                    pass
            if self._VLC:
                try:
                    self._VLC.release()
                except Exception:
                    pass
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Close -> {E}")
            
    def __del__(self):
        try:
            self.Close()
        except:
            pass


# IMPORT LIBRARIES
import tkinter as TK
import threading, math, time
from .N_GUI import GUI
from .N_Frame import Frame
from .N_Custom import Event_Bind
from .N_Custom import Event_Bind_Canvas
from .N_Canvas_Line import Canvas_Line
from .N_Canvas_Polyline import Canvas_Polyline
from .N_Canvas_Pie import Canvas_Pie
from .N_Canvas_Arc import Canvas_Arc
from .N_Canvas_Circle import Canvas_Circle
from .N_Canvas_Oval import Canvas_Oval
from .N_Canvas_Rectangle import Canvas_Rectangle
from .N_Canvas_RectangleR import Canvas_RectangleR
from .N_Canvas_Polygon import Canvas_Polygon
from .N_Canvas_Image import Canvas_Image
from .N_Canvas_Text import Canvas_Text
from .N_Canvas_Text_Old import Canvas_Text_Old

class Scroll:

    def __init__(self, Main, *args, **kwargs):
        self._GUI = GUI._Instance
        if self._GUI is not None:
            self._Type = "Scroll"
            try:
                self._Config = ['Name', 'Auto_Dark', 'Background', 'Light_Background', 'Dark_Background', 'Border_Color', 'Light_Border_Color', 'Dark_Border_Color', 'Border_Size', 'Resize', 'Popup', 'Left', 'Top', 'Width', 'Height', 'Scrollbar_Size', 'Vertical', 'Horizontal', 'Last']
                self._Initialized = False
                self._Name = False
                self._Last_Name = False
                self._Widget = []
                self._Item = []
                self._Resize = True
                self._Popup = False
                self._Display = True
                self._Main = Main
                self._Frame_Canvas = Frame(self._Main)
                self._Canvas_Scroll = TK.Canvas(self._Frame_Canvas._Frame)
                self._Scrollbar_Vertical = TK.Scrollbar(self._Frame_Canvas._Frame)
                self._Scrollbar_Horizontal = TK.Scrollbar(self._Frame_Canvas._Frame)
                self._Frame = TK.Canvas(self._Canvas_Scroll)
                self._Frame_Scroll = False
                self._Border_Color = '#000000'
                self._Border_Size = 0
                self._Background = self._Main._Background
                self._Background_Main = True
                self._Scrollbar_Size = 20
                self._Vertical = False
                self._Horizontal = False
                self._Width_Frame = 0
                self._Height_Frame = 0
                self._Last = False
                self._On_Resize = False
                self._Auto_Dark = True
                self._On_Show = False
                self._On_Hide = False
                self._Configure_After_Id = None
                self._Last_Width = 0
                self._Last_Height = 0
            except Exception as E:
                self._GUI.Error(f"{self._Type} -> Init -> {E}")
        else:
            print("Error: Gluonix -> GUI Instance Has Not Been Created")

    def __str__(self):
        return "Nucleon_Glunoix_Scroll[]"

    def __repr__(self):
        return "Nucleon_Glunoix_Scroll[]"
    
    def Copy(self, Name=False, Main=False):
        try:
            if not Main:
                Main = self._Main
            Instance = type(self)(Main)
            for Key in self._Config:
                if hasattr(self, "_"+Key):
                    setattr(Instance, "_"+Key, getattr(self, "_"+Key))
            if Name:
                setattr(Instance, "_Name", Name)
            Instance.Create()
            for Each in self._Widget:
                Each.Copy(Main=Instance)
            return Instance
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Copy -> {E}")
        
    def Delete(self):
        try:
            self._Main._Widget.remove(self)
            self._Frame.destroy()
            self._Scrollbar_Vertical.destroy()
            self._Scrollbar_Horizontal.destroy()
            self._Canvas_Scroll.destroy()
            self._Frame_Canvas.Delete()
            if self:
                del self
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete -> {E}")

    def Clear(self):
        try:
            for Each in list(self._Widget):
                Each.Delete()
            for Each in list(self._Item):
                Each.Delete()
            for Each in self._Frame.winfo_children():
                Each.destroy()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Clear -> {E}")

    def Refresh(self):
        try:
            self._Frame.update_idletasks()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Refresh -> {E}")
            
    def On_Configure(self, Event):
        try:
            if Event.width!=self._Last_Width or Event.height!=self._Last_Height:
                self._Last_Width = Event.width
                self._Last_Height = Event.height
                if self._Configure_After_Id is not None:
                    self._Canvas_Scroll.after_cancel(self._Configure_After_Id)
                self._Configure_After_Id = self._Canvas_Scroll.after(50, self.On_Configure_Debounced)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Configure -> {E}")
                
    def On_Configure_Debounced(self):
        try:
            self._Configure_After_Id = None
            T = threading.Thread(target=self.On_Configure_Thread, daemon=True)
            T.start()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Configure_Debounced -> {E}")
    
    def On_Configure_Thread(self):
        try:
            if self._Last:
                self.Update(self._Last)
            else:
                self.Update_All()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> On_Configure_Thread -> {E}")
            
    def Hide(self):
        try:
            self._Frame_Canvas.Hide()
            self._Display = False
            if self._On_Hide:
                self._On_Hide()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Hide -> {E}")
            
    def Show(self):
        try:
            self._Display = True
            self._Frame_Canvas.Show()
            if self._Last:
                self.Update(self._Last)
            else:
                self.Update_All()
            if self._On_Show:
                self._On_Show()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Show -> {E}")
            
    def Animate(self, Hide=False, Thread=True):
        try:
            self._Frame_Canvas.Animate(Hide=Hide, Thread=True)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate -> {E}")
            
    def Animate_Cancel(self):
        try:
            self._Frame_Canvas.Animate_Cancel()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate_Cancel -> {E}")
            
    def Animate_From(self, Left, Top, Time=None, Speed=None, Ease=None):
        try:
            if Time is not None:
                self._Animate_Time = Time
            if Speed is not None:
                self._Animate_Speed = Speed
            if Ease is not None:
                self._Animate_Ease = Ease
            self._Animate_Left, self._Animate_Top = Left, Top
            if Time is not None:
                self._Frame_Canvas._Animate_Time = Time
            if Speed is not None:
                self._Frame_Canvas._Animate_Speed = Speed
            if Ease is not None:
                self._Frame_Canvas._Animate_Ease = Ease
            self._Frame_Canvas._Animate_Left, self._Frame_Canvas._Animate_Top = Left, Top
            self._Frame_Canvas.Animate()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Animate_From -> {E}")
    
    def Grab(self, Path=False):
        try:
            return self._GUI.Grab_Widget(Path=Path, Widget=self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Grab -> {E}")
            
    def Top(self):
        try:
            self._Canvas_Scroll.yview_moveto(0)
            self._Canvas_Scroll.xview_moveto(0)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Top -> {E}")
            
    def Reset(self):
        try:
            self.Top()
            self._Width_Frame = self._Width
            self._Height_Frame = self._Height
            self._Frame.config(width=self._Width_Frame, height=self._Height_Frame)
            self._Scrollbar_Horizontal.place_forget()
            self._Scrollbar_Vertical.place_forget()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Reset -> {E}")
            
    def Scroll(self, E):
        try:
            if E.delta:
                self._Canvas_Scroll.yview_scroll(int(-1*(E.delta/120)), "units")
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Scroll -> {E}")
            
    def Widget(self):
        try:
            return self._Frame
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Widget -> {E}")
            
    def Bind(self, **Input):
        try:
            if 'On_Show' in Input:
                self._On_Show = Input['On_Show']
            if 'On_Hide' in Input:
                self._On_Hide = Input['On_Hide']
            Event_Bind(self._Frame, **Input)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind -> {E}")
            
    def Bind_Item(self, Item, **Input):
        try:
            Event_Bind_Canvas(self._Frame, Item, **Input)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Bind_Item -> {E}")
            
    def Config_Get(self, *Input):
        try:
            Return = {}
            for Each in self._Config:
                if Each in Input:
                    Return[Each] = getattr(self, "_"+Each)
            return Return
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config_Get -> {E}")
                
    def Config(self, **Input):
        try:
            Run = False
            for Each in self._Config:
                if Each in Input:
                    Value = Input[Each]
                    setattr(self, "_"+Each, Value)
                    Run = True
            self._Frame_Canvas.Config(**Input)
            if self._Initialized and Run:
                self.Create()
            if "Background" in Input:
                self._Background_Main = not bool(Input["Background"])
                for Each in self._Widget:
                    try:
                        if Each._Background_Main:
                            Each.Config(Background=False)
                        if hasattr(Each, '_Radius'):
                            if Each._Radius>0:
                                Each.Config(Radius=Each._Radius)
                    except Exception:
                        self.Nothing = False
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Config -> {E}")
            
    def Move(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left += Left
            if Top is not None:
                self._Top += Top
            if Left is not None or Top is not None:
                self.Position(Left=self._Left, Top=self._Top)
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Move -> {E}")
            
    def Center(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left-self._Width/2
            if Top is not None:
                self._Top = Top-self._Height/2
            if Left is not None or Top is not None:
                self.Position(Left=self._Left, Top=self._Top)
            return [self._Left+self._Width/2, self._Top+self._Height/2]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Center -> {E}")
            
    def Position(self, Left=None, Top=None):
        try:
            if Left is not None:
                self._Left = Left
            if Top is not None:
                self._Top = Top
            if Left is not None or Top is not None:
                self.Config(Left=self._Left, Top=self._Top)
            return self._Frame_Canvas.Position()
        except Exception as E:
            self._GUI.Error(f("{self._Type} -> Position -> {E}"))
            
    def Size(self, Width=False, Height=False):
        try:
            if Width:
                self._Width = Width
            if Height:
                self._Height = Height
            if Width or Height:
                self.Config(Width=self._Width, Height=self._Height)
            return self._Frame_Canvas.Size()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Size -> {E}")
    
    def Enlarge(self, Value=None):
        try:
            if Value is not None:
                self._Left -= Value
                self._Top -= Value
                self._Width += Value
                self._Height += Value
                self.Config(Left=self._Left, Top=self._Top, Width=self._Width, Height=self._Height)
                self.Create()
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Enlarge -> {E}")
    
    def Shrink(self, Value=None):
        try:
            if Value is not None:
                self._Left += Value
                self._Top += Value
                self._Width -= Value
                self._Height -= Value
                self.Config(Left=self._Left, Top=self._Top, Width=self._Width, Height=self._Height)
                self.Create()
            return True
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Shrink -> {E}")
            
    def Box(self):
        try:
            return self._Frame_Canvas.Box()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Box -> {E}")
        
    def Ratio(self):
        try:
            return self._Frame_Canvas.Ratio()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Ratio -> {E}")
        
    def Locate(self, Width, Height, Left, Top):
        try:
            Width = self._Width*(Width/100)
            Height = self._Height*(Height/100)
            Left = self._Width*(Left/100)-self._Border_Size
            Top = self._Height*(Top/100)-self._Border_Size
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate -> {E}")
        
    def Locate_Reverse(self, Width, Height, Left, Top):
        try:
            Width = round((Width/self._Width)*100, 3)
            Height = round((Height/self._Height)*100, 3)
            Left =  round((Left/self._Width)*100, 3)
            Top =  round((Top/self._Height)*100, 3)
            return [Width, Height, Left, Top]
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Locate_Reverse -> {E}")
            
    def Create(self):
        try:
            if not self._Background:
                self._Background = self._Main._Background
                if not hasattr(self, "_Light_Background"):
                    setattr(self, "_Light_Background", self._Background)
                if not hasattr(self, "_Dark_Background"):
                    setattr(self, "_Dark_Background", self._GUI.Invert(self._Background))
            if self._Auto_Dark and not self._GUI._Dark_Mode:
                self.Update_Color()
            if not self._Initialized:
                self.Update_Color()
                self._Width_Frame = self._Width
                self._Height_Frame = self._Height
                self._Frame_Canvas.Config(Width=self._Width, Height=self._Height, Left=self._Left, Top=self._Top)
                self._Frame_Canvas.Config(Background=self._Background, Border_Size=self._Border_Size, Border_Color=self._Border_Color)
                self._Frame_Canvas.Create()
                self._Canvas_Scroll.bind("<Configure>", self.On_Configure, add="+")
                self._Frame.bind("<MouseWheel>", self.Scroll)
                self._Frame.bind("<Configure>", lambda e: self._Canvas_Scroll.configure(scrollregion=self._Canvas_Scroll.bbox("all")))
                if not self._Display:
                    self.Hide()
                self._Main._Widget.append(self)
                self._Canvas_Scroll.config(background=self._Background, highlightthickness=0)
                self._Canvas_Scroll.place(relx=0, rely=0, relwidth=1, relheight=1)
                self._Frame.config(background=self._Background, highlightthickness=0)
                self._Initialized = True
            self._Scrollbar_Vertical.config(orient="vertical", command=self._Canvas_Scroll.yview, width=self._Scrollbar_Size)
            self._Scrollbar_Horizontal.config(orient="horizontal", command=self._Canvas_Scroll.xview, width=self._Scrollbar_Size)
            self._Frame.place(relx=0, rely=0, relwidth=(self._Width-self._Scrollbar_Size)/self._Width, relheight=(self._Height-self._Scrollbar_Size)/self._Height)
            self._Frame['highlightthickness']=0
            if not self._Frame_Scroll:
                self._Frame_Scroll = self._Canvas_Scroll.create_window((0, 0), window=self._Frame, anchor="nw")
                self._Canvas_Scroll.configure(yscrollcommand=self._Scrollbar_Vertical.set)
                self._Canvas_Scroll.configure(xscrollcommand=self._Scrollbar_Horizontal.set)
                self._Canvas_Scroll.pack(side="left", fill="both", expand=True)
            if self._Last:
                self.Update(self._Last)
            else:
                self.Update_All()
            if self._Vertical:
                self._Scrollbar_Vertical.place(relx=(self._Width-self._Scrollbar_Size)/self._Width, rely=0, relwidth=self._Scrollbar_Size/self._Width, relheight=1)
            if self._Horizontal:
                self._Scrollbar_Horizontal.place(relx=0, rely=(self._Height-self._Scrollbar_Size)/self._Height, relwidth=1, relheight=self._Scrollbar_Size/self._Height)
            if self._Name!=self._Last_Name:
                if self._Last_Name:
                    if self._Last_Name in self._Main.__dict__:
                        del self._Main.__dict__[self._Last_Name]
                if self._Name:
                    self._Main.__dict__[self._Name] = self
                self._Last_Name = self._Name
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Create -> {E}")
            
    def Update_Color(self):
        try:
            self._GUI.Initiate_Colors(self)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Update_Color -> {E}")
             
    def _Update_Scrollbars(self):
        try:
            Left, Top, Width, Height = self._Frame_Canvas.Box()
            self._Frame.config(width=self._Width_Frame, height=self._Height_Frame)
            if self._Vertical or Height < self._Height_Frame:
                self._Scrollbar_Vertical.place(relx=(self._Width-self._Scrollbar_Size)/self._Width, rely=0, relwidth=self._Scrollbar_Size/self._Width, relheight=1)
            else:
                self._Scrollbar_Vertical.place_forget()
            if self._Horizontal or Width < self._Width_Frame:
                self._Scrollbar_Horizontal.place(relx=0, rely=(self._Height-self._Scrollbar_Size)/self._Height, relwidth=1, relheight=self._Scrollbar_Size/self._Height)
            else:
                self._Scrollbar_Horizontal.place_forget()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> _Update_Scrollbars -> {E}")
             
    def Update(self, Widget):
        try:
            Extra_Space = self._Scrollbar_Size
            Left, Top, Width, Height = Widget.Box()
            Right = Left + Width
            Bottom = Top + Height
            if self._Width_Frame < Right:
                self._Width_Frame = Right + Extra_Space*2
            if self._Height_Frame < Bottom:
                self._Height_Frame = Bottom + Extra_Space*3
            self._Update_Scrollbars()
            self._Last = Widget
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Update -> {E}")
             
    def Update_All(self):
        try:
            Extra_Space = self._Scrollbar_Size
            for Widget in self._Widget:
                Left, Top, Width, Height = Widget.Box()
                Right = Left + Width
                Bottom = Top + Height
                if self._Width_Frame < Right:
                    self._Width_Frame = Right + Extra_Space
                if self._Height_Frame < Bottom:
                    self._Height_Frame = Bottom + Extra_Space
            self._Update_Scrollbars()
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Update_All -> {E}")

    def Hide_Item(self, Item=False):
        try:
            if Item:
                self._Frame.itemconfigure(Item, state='hidden')
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Hide_Item -> {E}")

    def Show_Item(self, Item=False):
        try:
            if Item:
                self._Frame.itemconfigure(Item, state='normal')
                self._Frame.tag_raise(Item)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Show_Item -> {E}")
    
    def Delete_Item(self, Item=False):
        try:
            if Item:
                self._Frame.delete(Item)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete_Item -> {E}")
    
    def Delete_All(self):
        try:
            self._Frame.delete('all')
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Delete_All -> {E}")
            
    def Find_Near(self, X, Y):
        try:
            return self._Frame.find_closest(X, Y)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Find_Near -> {E}")
            
    def Find_Overlap(self, X1, Y1, X2, Y2):
        try:
            return self._Frame.find_overlapping(X1, Y1, X2, Y2)
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Find_Overlap -> {E}")
            
    def Line(self, Name=False):
        try:
            Item = Canvas_Line(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Line -> {E}")
            
    def Polyline(self, Name=False):
        try:
            Item = Canvas_Polyline(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Polyline -> {E}")
                
    def Pie(self, Name=False):
        try:
            Item = Canvas_Pie(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Pie -> {E}")
                
    def Arc(self, Name=False):
        try:
            Item = Canvas_Arc(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Arc -> {E}")
                
    def Circle(self, Name=False):
        try:
            Item = Canvas_Circle(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Circle -> {E}")
                
    def Oval(self, Name=False):
        try:
            Item = Canvas_Oval(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Oval -> {E}")
            
    def Polygon(self, Name=False):
        try:
            Item = Canvas_Polygon(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Polygon -> {E}")
                
    def Rectangle(self, Name=False):
        try:
            Item = Canvas_Rectangle(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Rectangle -> {E}")
                
    def RectangleR(self, Name=False):
        try:
            Item = Canvas_RectangleR(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> RectangleR -> {E}")
                
    def Image(self, Name=False):
        try:
            Item = Canvas_Image(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Image -> {E}")
            
    def Text(self, Name=False):
        try:
            Item = Canvas_Text(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Text -> {E}")
            
    def Text_Old(self, Name=False):
        try:
            Item = Canvas_Text_Old(self)
            if Name:
                Item.Config(Name=Name)
            return Item
        except Exception as E:
            self._GUI.Error(f"{self._Type} -> Text_Old -> {E}")
